from __future__ import annotations

import os
from collections import OrderedDict
from collections.abc import Generator
from io import BytesIO
from pathlib import PurePosixPath
from typing import IO, TypeVar

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.workspace import ObjectInfo
from databricks.labs.blueprint.paths import WorkspacePath

from databricks.labs.ucx.source_code.base import decode_with_bom


# lru_cache won't let us invalidate cache entries
# so we provide our own custom lru_cache
class _PathLruCache:

    _datas: OrderedDict[PurePosixPath, bytes]
    """Cached binary data of files, keyed by workspace path, ordered from oldest to newest."""

    _max_entries: int
    """The maximum number of entries to hold in the cache."""

    def __init__(self, max_entries: int) -> None:
        # Ordered from oldest to newest.
        self._datas = OrderedDict()
        self._max_entries = max_entries

    @classmethod
    def _normalize(cls, path: _CachedPath) -> PurePosixPath:
        # Note: must not return the same instance that was passed in, to avoid circular references (and memory leaks).
        return PurePosixPath(*path.parts)

    def load(self, cached_path: _CachedPath, buffering: int = -1) -> bytes:
        normalized_path = self._normalize(cached_path)

        data = self._datas.get(normalized_path, None)
        if data is not None:
            self._datas.move_to_end(normalized_path)
            return data

        # Need to bypass the _CachedPath.open() override to actually open and retrieve the file content.
        with WorkspacePath.open(cached_path, mode="rb", buffering=buffering) as workspace_file:
            data = workspace_file.read()
        if self._max_entries <= len(self._datas):
            self._datas.popitem(last=False)
        self._datas[normalized_path] = data
        return data

    def clear(self) -> None:
        self._datas.clear()

    def remove(self, path: _CachedPath) -> None:
        del self._datas[self._normalize(path)]


class _CachedPath(WorkspacePath):
    def __init__(self, cache: _PathLruCache, ws: WorkspaceClient, *args: str | bytes | os.PathLike) -> None:
        super().__init__(ws, *args)
        self._cache = cache

    @classmethod
    def _from_object_info_with_cache(
        cls,
        cache: _PathLruCache,
        ws: WorkspaceClient,
        object_info: ObjectInfo,
    ) -> _CachedPath:
        assert object_info.path
        path = cls(cache, ws, object_info.path)
        path._cached_object_info = object_info
        return path

    def with_segments(self: _CachedPathT, *path_segments: bytes | str | os.PathLike) -> _CachedPathT:
        return type(self)(self._cache, self._ws, *path_segments)

    def iterdir(self) -> Generator[_CachedPath, None, None]:
        # Variant of the superclass implementation that preserves the cache, as well as the client.
        for object_info in self._ws.workspace.list(self.as_posix()):
            yield self._from_object_info_with_cache(self._cache, self._ws, object_info)

    def open(  # type: ignore[override]
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        # We only cache reads; if a write happens we use the default implementation (and evict any cache entry).
        if 'w' in mode:
            self._cache.remove(self)
            return super().open(mode, buffering, encoding, errors, newline)

        binary_data = self._cache.load(self, buffering=buffering)
        binary_io = BytesIO(binary_data)
        if 'b' in mode:
            return binary_io

        return decode_with_bom(binary_io, encoding, errors, newline)

    # _rename calls unlink so no need to override it
    def unlink(self, missing_ok: bool = False) -> None:
        self._cache.remove(self)
        return super().unlink(missing_ok)


_CachedPathT = TypeVar("_CachedPathT", bound=_CachedPath)


class InvalidPath(ValueError):
    pass


class WorkspaceCache:

    def __init__(self, ws: WorkspaceClient, max_entries: int = 2048) -> None:
        self._ws = ws
        self._cache = _PathLruCache(max_entries)

    def get_workspace_path(self, path: str) -> WorkspacePath:
        """Obtain a `WorkspacePath` instance for a path that refers to a workspace file or notebook.

        The instance returned participates in this content cache: the first time the path is opened the content will
        be immediately retrieved (prior to reading) and cached.

        Args:
            path: a valid workspace path (must be absolute)
        Raises:
            InvalidPath: this is raised immediately if the supplied path is not a syntactically
                valid workspace path. (This is not raised if the path is syntactically valid but does not exist.)
        """
        if not path.startswith("/"):
            msg = f"Invalid workspace path; must be absolute and start with a slash ('/'): {path}"
            raise InvalidPath(msg)
        return _CachedPath(self._cache, self._ws, path)



from __future__ import annotations

import logging
from collections.abc import Sequence, Iterable

from databricks.labs.blueprint.paths import WorkspacePath
from databricks.sdk import WorkspaceClient

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import DatabricksError, NotFound

from databricks.labs.ucx.framework.owners import (
    Ownership,
    AdministratorLocator,
    WorkspacePathOwnership,
    LegacyQueryOwnership,
)
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.source_code.base import DirectFsAccess

logger = logging.getLogger(__name__)


class DirectFsAccessCrawler(CrawlerBase[DirectFsAccess]):

    @classmethod
    def for_paths(cls, backend: SqlBackend, schema) -> DirectFsAccessCrawler:
        return DirectFsAccessCrawler(backend, schema, "directfs_in_paths")

    @classmethod
    def for_queries(cls, backend: SqlBackend, schema) -> DirectFsAccessCrawler:
        return DirectFsAccessCrawler(backend, schema, "directfs_in_queries")

    def __init__(self, sql_backend: SqlBackend, schema: str, table: str):
        """
        Initializes a DFSACrawler instance.

        Args:
            sql_backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema: The schema name for the inventory persistence.
        """
        super().__init__(
            sql_backend=sql_backend, catalog="hive_metastore", schema=schema, table=table, klass=DirectFsAccess
        )

    def dump_all(self, dfsas: Sequence[DirectFsAccess]) -> None:
        """This crawler doesn't follow the pull model because the fetcher fetches data for 2 crawlers, not just one
        It's not **bad** because all records are pushed at once.
        Providing a multi-entity crawler is out-of-scope of this PR
        """
        try:
            # TODO: Until we historize data, we append all DFSAs
            # UPDATE: We historize DFSA from WorkflowLinter, not from QueryLinter yet
            self._update_snapshot(dfsas, mode="append")
        except DatabricksError as e:
            logger.error("Failed to store DFSAs", exc_info=e)

    def _try_fetch(self) -> Iterable[DirectFsAccess]:
        sql = f"SELECT * FROM {escape_sql_identifier(self.full_name)}"
        for row in self._sql_backend.fetch(sql):
            yield self._klass.from_dict(row.asDict())

    def _crawl(self) -> Iterable[DirectFsAccess]:
        return []
        # TODO raise NotImplementedError() once CrawlerBase supports empty snapshots


class DirectFsAccessOwnership(Ownership[DirectFsAccess]):
    """Determine ownership of records reporting direct filesystem access.

    This is intended to be:

     - For queries, the creator of the query (if known).
     - For jobs, the owner of the path for the notebook or source (if known).
    """

    def __init__(
        self,
        administrator_locator: AdministratorLocator,
        workspace_path_ownership: WorkspacePathOwnership,
        legacy_query_ownership: LegacyQueryOwnership,
        workspace_client: WorkspaceClient,
    ) -> None:
        super().__init__(administrator_locator)
        self._workspace_path_ownership = workspace_path_ownership
        self._legacy_query_ownership = legacy_query_ownership
        self._workspace_client = workspace_client

    def _maybe_direct_owner(self, record: DirectFsAccess) -> str | None:
        if record.source_type == 'QUERY' and record.query_id:
            return self._legacy_query_ownership.owner_of(record.query_id)
        if record.source_type in {'NOTEBOOK', 'FILE'}:
            return self._notebook_owner(record)
        logger.warning(f"Unknown source type {record.source_type} for {record.source_id}")
        return None

    def _notebook_owner(self, record):
        try:
            workspace_path = WorkspacePath(self._workspace_client, record.source_id)
            owner = self._workspace_path_ownership.owner_of(workspace_path)
            return owner
        except NotFound:
            return None



import collections
import logging
import re
from dataclasses import dataclass, replace
from functools import cached_property
from typing import ClassVar
from packaging.version import Version, InvalidVersion


from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import AlreadyExists, NotFound, BadRequest
from databricks.sdk.service.catalog import (
    ConnectionType,
    ConnectionInfo,
    SecurableType,
    PermissionsChange,
    CatalogInfo,
    Privilege,
)

from databricks.labs.ucx.account.workspaces import WorkspaceInfo
from databricks.labs.ucx.assessment.secrets import SecretsMixin
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.hive_metastore import ExternalLocations


logger = logging.getLogger(__name__)


@dataclass
class ExternalHmsInfo:
    """
    This is a dataclass that represents the external Hive Metastore connection information.
    It supports non glue external metastores.
    """

    database_type: str
    host: str
    port: str
    database: str
    user: str | None
    password: str | None
    version: str | None

    def as_dict(self) -> dict[str, str]:
        return {
            "database": self.database,
            "db_type": self.database_type,
            "host": self.host,
            "port": self.port,
        }


class HiveMetastoreFederationEnabler:
    def __init__(self, installation: Installation):
        self._installation = installation

    def enable(self):
        config = self._installation.load(WorkspaceConfig)
        config.enable_hms_federation = True
        self._installation.save(config)


class HiveMetastoreFederation(SecretsMixin):
    def __init__(
        self,
        ws: WorkspaceClient,
        external_locations: ExternalLocations,
        workspace_info: WorkspaceInfo,
        config: WorkspaceConfig,
        *,
        enable_hms_federation: bool = False,
    ):
        self._ws = ws
        self._external_locations = external_locations
        self._workspace_info = workspace_info
        self._enable_hms_federation = enable_hms_federation
        self._config = config

    # Supported databases and version for HMS Federation
    supported_database_versions: ClassVar[dict[str, list[str]]] = {
        "mysql": ["2.3", "0.13"],
    }

    def create_from_cli(self, prompts: Prompts) -> None:
        if not self._enable_hms_federation:
            raise RuntimeWarning('Run `databricks labs ucx enable-hms-federation` to enable HMS Federation')

        name = prompts.question(
            'Enter the name of the Hive Metastore connection and catalog', default=self._workspace_info.current()
        )

        if self._external_hms and prompts.confirm(
            f'A supported external Hive Metastore connection was identified: {self._external_hms.database_type}. '
            f'Use this connection?'
        ):
            connection_info = self._get_or_create_ext_connection(name, self._external_hms)
        else:
            connection_info = self._get_or_create_int_connection(name)

        assert connection_info.name is not None
        self._register_federated_catalog(connection_info)

    @cached_property
    def _external_hms(self) -> ExternalHmsInfo | None:
        if not self._config.spark_conf:
            logger.info('Spark config not found')
            return None
        spark_config = self._config.spark_conf
        jdbc_url = self._get_value_from_config_key(spark_config, 'spark.hadoop.javax.jdo.option.ConnectionURL')
        if not jdbc_url:
            logger.info('JDBC URL not found')
            return None
        version_value = self._get_value_from_config_key(spark_config, 'spark.sql.hive.metastore.version')
        if not version_value:
            logger.info('Hive Metastore version not found')
            return None
        try:
            version = Version(version_value)
        except InvalidVersion:
            logger.info('Hive Metastore version is not valid')
            return None
        major_minor_version = f"{version.major}.{version.minor}"
        external_hms = replace(self._split_jdbc_url(jdbc_url), version=major_minor_version)
        supported_versions = self.supported_database_versions.get(external_hms.database_type)
        if not supported_versions:
            logger.info(f'Unsupported Hive Metastore: {external_hms.database_type}')
            return None
        if major_minor_version not in supported_versions:
            logger.info(f'Unsupported Hive Metastore Version: {external_hms.database_type} - {version}')
            return None

        if not external_hms.user:
            external_hms = replace(
                external_hms,
                user=self._get_value_from_config_key(spark_config, 'spark.hadoop.javax.jdo.option.ConnectionUserName'),
            )
        if not external_hms.password:
            external_hms = replace(
                external_hms,
                password=self._get_value_from_config_key(
                    spark_config, 'spark.hadoop.javax.jdo.option.ConnectionPassword'
                ),
            )
        return external_hms

    @classmethod
    def _split_jdbc_url(cls, jdbc_url: str) -> ExternalHmsInfo:
        # Define the regex pattern to match the JDBC URL components
        pattern = re.compile(
            r'jdbc:(?P<db_type>[a-zA-Z0-9]+)://(?P<host>[^:/]+):(?P<port>\d+)/(?P<database>[^?]+)(\?user=(?P<user>[^&]+)&password=(?P<password>[^&]+))?'
        )
        match = pattern.match(jdbc_url)
        if not match:
            raise ValueError(f'Unsupported JDBC URL: {jdbc_url}')

        db_type = match.group('db_type')
        host = match.group('host')
        port = match.group('port')
        database = match.group('database')
        user = match.group('user')
        password = match.group('password')

        return ExternalHmsInfo(db_type, host, port, database, user, password, None)

    def _register_federated_catalog(
        self,
        connection_info,
    ) -> CatalogInfo:
        try:
            return self._ws.catalogs.create(
                name=connection_info.name,
                connection_name=connection_info.name,
                options={"authorized_paths": self._get_authorized_paths()},
            )
        except BadRequest as err:
            if err.error_code == 'CATALOG_ALREADY_EXISTS':
                logger.info(f'Catalog {connection_info.name} already exists')
                for catalog_info in self._ws.catalogs.list():
                    if catalog_info.name == connection_info.name:
                        return catalog_info
            raise err

    def _get_or_create_int_connection(self, name: str) -> ConnectionInfo:
        try:
            return self._ws.connections.create(
                name=name,
                connection_type=ConnectionType.HIVE_METASTORE,  # needs SDK change
                options={"builtin": "true"},
            )
        except AlreadyExists:
            return self._get_existing_connection(name)

    def _get_existing_connection(self, name: str) -> ConnectionInfo:
        for connection in self._ws.connections.list():
            if connection.name == name:
                return connection
        raise NotFound(f'Connection {name} not found')

    def _get_or_create_ext_connection(self, name: str, external_hms: ExternalHmsInfo) -> ConnectionInfo:
        options = external_hms.as_dict()
        if external_hms.user:
            options["user"] = external_hms.user
        if external_hms.password:
            options["password"] = external_hms.password
        if external_hms.version:
            options["version"] = external_hms.version
        try:
            return self._ws.connections.create(
                name=name,
                connection_type=ConnectionType.HIVE_METASTORE,  # needs SDK change
                options=options,
            )
        except AlreadyExists:
            return self._get_existing_connection(name)

    def _get_authorized_paths(self) -> str:
        existing = {}
        for external_location in self._ws.external_locations.list():
            existing[external_location.url] = external_location
        authorized_paths = []
        current_user = self._ws.current_user.me()
        if not current_user.user_name:
            raise NotFound('Current user not found')
        # Get the external locations. If not using external HMS, include the root DBFS location.
        if self._external_hms is not None:
            external_locations = self._external_locations.external_locations_with_root()
        else:
            external_locations = self._external_locations.snapshot()

        for external_location_info in external_locations:
            location = ExternalLocations.clean_location(external_location_info.location)
            existing_location = existing.get(location)
            if not existing_location:
                logger.warning(f'External location {location} not found')
                continue
            location_name = existing_location.name
            if not location_name:
                logger.warning(f'External location {location} has no name')
                continue
            self._add_missing_permissions_if_needed(location_name, current_user.user_name)
            authorized_paths.append(location)
        return ",".join(authorized_paths)

    def _add_missing_permissions_if_needed(self, location_name: str, current_user: str):
        grants = self._location_grants(location_name)
        if Privilege.CREATE_FOREIGN_SECURABLE not in grants[current_user]:
            change = PermissionsChange(principal=current_user, add=[Privilege.CREATE_FOREIGN_SECURABLE])
            self._ws.grants.update(SecurableType.EXTERNAL_LOCATION, location_name, changes=[change])

    def _location_grants(self, location_name: str) -> dict[str, set[Privilege]]:
        grants: dict[str, set[Privilege]] = collections.defaultdict(set)
        result = self._ws.grants.get(SecurableType.EXTERNAL_LOCATION, location_name)
        if not result.privilege_assignments:
            return grants
        for assignment in result.privilege_assignments:
            if not assignment.privileges:
                continue
            if not assignment.principal:
                continue
            for privilege in assignment.privileges:
                grants[assignment.principal].add(privilege)
        return grants



import base64
import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass
from typing import ClassVar

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, BadRequest
from databricks.sdk.service.compute import (
    ClusterDetails,
    ClusterSource,
    DataSecurityMode,
    DbfsStorageInfo,
    InitScriptInfo,
    LocalFileInfo,
    Policy,
    WorkspaceStorageInfo,
)

from databricks.labs.ucx.assessment.crawlers import (
    AZURE_SP_CONF_FAILURE_MSG,
    INCOMPATIBLE_SPARK_CONFIG_KEYS,
    INIT_SCRIPT_DBFS_PATH,
    INIT_SCRIPT_LOCAL_PATH,
    azure_sp_conf_present_check,
    spark_version_compatibility,
    is_mlr,
)
from databricks.labs.ucx.assessment.init_scripts import CheckInitScriptMixin
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.owners import Ownership
from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


@dataclass
class ClusterInfo:
    cluster_id: str
    success: int
    failures: str
    spark_version: str | None = None
    policy_id: str | None = None
    cluster_name: str | None = None
    creator: str | None = None
    """User-name of the creator of the cluster, if known."""

    __id_attributes__: ClassVar[tuple[str, ...]] = ("cluster_id",)

    @classmethod
    def from_cluster_details(cls, details: ClusterDetails):
        return ClusterInfo(
            cluster_id=details.cluster_id if details.cluster_id else "",
            cluster_name=details.cluster_name,
            policy_id=details.policy_id,
            spark_version=details.spark_version,
            creator=details.creator_user_name or None,
            success=1,
            failures="[]",
        )


class CheckClusterMixin(CheckInitScriptMixin):
    _ws: WorkspaceClient

    def _safe_get_cluster_policy(self, policy_id: str) -> Policy | None:
        try:
            return self._ws.cluster_policies.get(policy_id)
        except NotFound:
            logger.warning(f"The cluster policy was deleted: {policy_id}")
            return None

    def _check_cluster_policy(self, policy_id: str, source: str) -> list[str]:
        failures: list[str] = []
        policy = self._safe_get_cluster_policy(policy_id)
        if policy:
            if policy.definition:
                if azure_sp_conf_present_check(json.loads(policy.definition)):
                    failures.append(f"{AZURE_SP_CONF_FAILURE_MSG} {source}.")
            if policy.policy_family_definition_overrides:
                if azure_sp_conf_present_check(json.loads(policy.policy_family_definition_overrides)):
                    failures.append(f"{AZURE_SP_CONF_FAILURE_MSG} {source}.")
        return failures

    def _get_init_script_data(self, init_script_info: InitScriptInfo) -> str | None:
        try:
            match init_script_info:
                case InitScriptInfo(dbfs=DbfsStorageInfo(destination)):
                    split = destination.split(":")
                    if len(split) != INIT_SCRIPT_DBFS_PATH:
                        return None
                    data = self._ws.dbfs.read(split[1]).data
                    if data is not None:
                        return base64.b64decode(data).decode("utf-8")
                case InitScriptInfo(workspace=WorkspaceStorageInfo(workspace_file_destination)):
                    data = self._ws.workspace.export(workspace_file_destination).content
                    if data is not None:
                        return base64.b64decode(data).decode("utf-8")
                case InitScriptInfo(file=LocalFileInfo(destination)):
                    split = destination.split(":/")
                    if len(split) != INIT_SCRIPT_LOCAL_PATH:
                        return None
                    with open(split[1], "r", encoding="utf-8") as file:
                        data = file.read()
                    return data

            return None
        except (NotFound, BadRequest, FileNotFoundError, UnicodeDecodeError) as e:
            logger.warning(f"Error reading {init_script_info}: {e}")
            return None

    def _check_cluster_init_script(self, init_scripts: list[InitScriptInfo], source: str) -> list[str]:
        failures: list[str] = []
        for init_script_info in init_scripts:
            init_script_data = self._get_init_script_data(init_script_info)
            failures.extend(self.check_init_script(init_script_data, source))
        return failures

    def _check_spark_conf(self, conf: dict[str, str], source: str) -> list[str]:
        failures: list[str] = []
        for key, error in INCOMPATIBLE_SPARK_CONFIG_KEYS.items():
            if key in conf:
                failures.append(f"{error}: {key} in {source}.")
        for value in conf.values():
            if "dbfs:/mnt" in value or "/dbfs/mnt" in value:
                failures.append(f"using DBFS mount in configuration: {value}")
        # Checking if Azure cluster config is present in spark config
        if azure_sp_conf_present_check(conf):
            failures.append(f"{AZURE_SP_CONF_FAILURE_MSG} {source}.")
        return failures

    def _check_cluster_failures(self, cluster: ClusterDetails, source: str) -> list[str]:
        failures: list[str] = []

        unsupported_cluster_types = [
            DataSecurityMode.LEGACY_PASSTHROUGH,
            DataSecurityMode.LEGACY_SINGLE_USER,
            DataSecurityMode.LEGACY_TABLE_ACL,
        ]
        support_status = spark_version_compatibility(cluster.spark_version)
        if support_status != "supported":
            failures.append(f"not supported DBR: {cluster.spark_version}")
        if cluster.spark_conf is not None:
            failures.extend(self._check_spark_conf(cluster.spark_conf, source))
        # Checking if Azure cluster config is present in cluster policies
        if cluster.policy_id is not None:
            failures.extend(self._check_cluster_policy(cluster.policy_id, source))
        if cluster.init_scripts is not None:
            failures.extend(self._check_cluster_init_script(cluster.init_scripts, source))
        if cluster.data_security_mode == DataSecurityMode.NONE:
            failures.append("No isolation shared clusters not supported in UC")
        if cluster.data_security_mode in unsupported_cluster_types:
            failures.append(f"cluster type not supported : {cluster.data_security_mode.value}")
        if cluster.data_security_mode == DataSecurityMode.NONE and is_mlr(cluster.spark_version):
            failures.append("Shared Machine Learning Runtime clusters are not supported in UC")

        return failures


class ClustersCrawler(CrawlerBase[ClusterInfo], CheckClusterMixin):
    def __init__(self, ws: WorkspaceClient, sql_backend: SqlBackend, schema: str):
        super().__init__(sql_backend, "hive_metastore", schema, "clusters", ClusterInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[ClusterInfo]:
        all_clusters = list(self._ws.clusters.list())
        return list(self._assess_clusters(all_clusters))

    def _assess_clusters(self, all_clusters: Iterable[ClusterDetails]):
        for cluster in all_clusters:
            if cluster.cluster_source == ClusterSource.JOB:
                continue
            creator = cluster.creator_user_name or None
            if not creator:
                logger.warning(
                    f"Cluster {cluster.cluster_id} have Unknown creator, it means that the original creator "
                    f"has been deleted and should be re-created"
                )
            cluster_info = ClusterInfo.from_cluster_details(cluster)
            failures = self._check_cluster_failures(cluster, "cluster")
            if len(failures) > 0:
                cluster_info.success = 0
                cluster_info.failures = json.dumps(failures)
            yield cluster_info

    def _try_fetch(self) -> Iterable[ClusterInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield ClusterInfo(*row)


class ClusterOwnership(Ownership[ClusterInfo]):
    """Determine ownership of clusters in the inventory.

    This is the cluster creator (if known).
    """

    def _maybe_direct_owner(self, record: ClusterInfo) -> str | None:
        return record.creator


@dataclass
class PolicyInfo:
    policy_id: str
    policy_name: str
    success: int
    failures: str
    spark_version: str | None = None
    policy_description: str | None = None
    creator: str | None = None
    """User-name of the creator of the cluster policy, if known."""

    __id_attributes__: ClassVar[tuple[str, ...]] = ("policy_id",)


class PoliciesCrawler(CrawlerBase[PolicyInfo], CheckClusterMixin):
    def __init__(self, ws: WorkspaceClient, sql_backend: SqlBackend, schema):
        super().__init__(sql_backend, "hive_metastore", schema, "policies", PolicyInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[PolicyInfo]:
        all_policies = list(self._ws.cluster_policies.list())
        return list(self._assess_policies(all_policies))

    def _assess_policies(self, all_policies: Iterable[Policy]) -> Iterable[PolicyInfo]:
        for policy in all_policies:
            failures: list[str] = []
            if policy.policy_id is None:
                continue
            failures.extend(self._check_cluster_policy(policy.policy_id, "policy"))
            spark_version = None
            if policy.definition is not None:
                try:
                    spark_version = json.dumps(json.loads(policy.definition)["spark_version"])
                except KeyError:
                    pass
            policy_name = policy.name or "UNDEFINED"
            creator_name = policy.creator_user_name or None

            policy_info = PolicyInfo(
                policy_id=policy.policy_id,
                policy_description=policy.description,
                policy_name=policy_name,
                spark_version=spark_version,
                success=1,
                failures="[]",
                creator=creator_name,
            )
            if len(failures) > 0:
                policy_info.success = 0
                policy_info.failures = json.dumps(failures)
            yield policy_info

    def _try_fetch(self) -> Iterable[PolicyInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield PolicyInfo(*row)


class ClusterPolicyOwnership(Ownership[PolicyInfo]):
    """Determine ownership of cluster policies in the inventory.

    This is the creator of the cluster policy (if known).
    """

    def _maybe_direct_owner(self, record: PolicyInfo) -> str | None:
        return record.creator



from __future__ import annotations
import dataclasses
import datetime as dt
import json
import logging
from enum import Enum, EnumMeta
from collections.abc import Iterable, Sequence
from typing import Any, ClassVar, Generic, Protocol, TypeVar, get_type_hints

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.framework.owners import Ownership
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.progress.install import Historical


logger = logging.getLogger(__name__)


class DataclassWithIdAttributes(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Any]]

    __id_attributes__: ClassVar[tuple[str, ...]]
    """The names of attributes (can be dataclass fields or ordinary properties) that make up the object identifier.

    All attributes must be (non-optional) strings.
    """


Record = TypeVar("Record", bound=DataclassWithIdAttributes)
T = TypeVar("T")


class HistoricalEncoder(Generic[Record]):
    """An encoder for dataclasses that will be stored in our history log.

    Our history records are designed with several goals in mind:
     - Records can be of different types, meaning we have to store heterogenous types using a homogenous schema.
     - We partially shred the records to allow for easier SQL-based querying.
     - Flexibility from the start, because schema changes in the future will be very difficult.
     - Records will be shared (and queried) across workspaces with different versions of UCX installed.

    With this in mind:
     - We have an `object-type` (discriminator) field for holding the type of the record we're storing. This allows for
       type-specific logic when querying.
     - We have a (type-specific) field for storing the business key (identifier) of each record.
     - We have special handling for the `failures` attribute that is often present on the types we are storing.
     - The `object-data` is an unconstrained key-value map, allowing arbitrary attributes to be stored and (if top-level)
       queried directly. Complex values are JSON-encoded.
     - The :py:class:`Ownership` mechanism is used to associate each record with a user.
     - To help with forward and backward compatibility the UCX version used to encode a record is included in each
       record.
     - The associated workspace and job run is also included in each record.
     - These records are currently stored in a (UC-based) table that is shared across workspaces attached to the same
       metastore. They are used to help report on migration progress.
    """

    _job_run_id: int
    """The identifier of the current job run, with which these records are associated."""

    _workspace_id: int
    """The identifier of the current workspace, for identifying where these records came from."""

    _ownership: Ownership[Record]
    """Used to determine the owner for each record."""

    _object_type: str
    """The name of the record class being encoded by this instance."""

    _field_names_with_types: dict[str, type]
    """A map of the fields on instances of the record class (and their types); these will appear in the object data."""

    _has_failures: type[str | list[str]] | None
    """The type of the failures attribute for this record, if present."""

    _id_attribute_names: Sequence[str]
    """The names of the record attributes that are used to produce the identifier for each record.

    Attributes can be either a dataclass field or a property, for which the type must be a (non-optional) string.
    """

    def __init__(self, job_run_id: int, workspace_id: int, ownership: Ownership[Record], klass: type[Record]) -> None:
        self._job_run_id = job_run_id
        self._workspace_id = workspace_id
        self._ownership = ownership
        self._object_type = self._get_object_type(klass)
        self._field_names_with_types, self._has_failures = self._get_field_names_with_types(klass)
        self._id_attribute_names = self._get_id_attribute_names(klass)

    @classmethod
    def _get_field_names_with_types(cls, klass: type[Record]) -> tuple[dict[str, type], type[str | list[str]] | None]:
        """Return the dataclass-defined fields that the record type declares, and their associated types.

        If the record has a "failures" attribute this is treated specially: it is removed but we signal that it was
        present.

        Arguments:
            klass: The record type.
        Returns:
            A tuple containing:
                - A dictionary of fields to include in the object data, and their type.
                - The type of the failures field, if present.
        """
        # Ignore the field types returned by dataclasses.fields(): it doesn't resolve string-based annotations (which
        # are produced automatically in a __future__.__annotations__ context). Unfortunately the dataclass mechanism
        # captures the type hints prior to resolution (which happens later in the class initialization process).
        # As such, we rely on dataclasses.fields() for the set of field names, but not the types which we fetch directly.
        klass_type_hints = get_type_hints(klass)
        field_names = [field.name for field in dataclasses.fields(klass)]
        field_names_with_types = {field_name: klass_type_hints[field_name] for field_name in field_names}
        if "failures" not in field_names_with_types:
            failures_type = None
        else:
            failures_type = field_names_with_types.pop("failures")
            if failures_type not in (str, list[str]):
                msg = f"Historical record {klass} has invalid 'failures' attribute of type: {failures_type}"
                raise TypeError(msg)
        return field_names_with_types, failures_type

    def _get_id_attribute_names(self, klazz: type[Record]) -> Sequence[str]:
        id_attribute_names = tuple(klazz.__id_attributes__)
        all_fields = self._field_names_with_types
        for name in id_attribute_names:
            id_attribute_type = all_fields.get(name, None) or self._detect_property_type(klazz, name)
            if id_attribute_type is None:
                raise AttributeError(name=name, obj=klazz)
            if id_attribute_type != str:
                msg = f"Historical record {klazz} has a non-string id attribute: {name} (type={id_attribute_type})"
                raise TypeError(msg)
        return id_attribute_names

    def _detect_property_type(self, klazz: type[Record], name: str) -> str | None:
        maybe_property = getattr(klazz, name, None)
        if maybe_property is None:
            return None
        if not isinstance(maybe_property, property):
            msg = f"Historical record {klazz} declares an id attribute that is not a field or property: {name} (type={maybe_property})"
            raise TypeError(msg)
        property_getter = maybe_property.fget
        if not property_getter:
            msg = f"Historical record {klazz} has a non-readable property as an id attribute: {name}"
            raise TypeError(msg)
        type_hints = get_type_hints(property_getter) or {}
        try:
            return type_hints["return"]
        except KeyError as e:
            msg = f"Historical record {klazz} has a property with no type as an id attribute: {name}"
            raise TypeError(msg) from e

    @classmethod
    def _get_object_type(cls, klass: type[Record]) -> str:
        return klass.__name__

    @classmethod
    def _as_dict(cls, record: Record) -> dict[str, Any]:
        return dataclasses.asdict(record)

    def _object_id(self, record: Record) -> list[str]:
        return [getattr(record, field) for field in self._id_attribute_names]

    @classmethod
    def _encode_non_serializable(cls, name: str, value: Any) -> Any:
        if isinstance(type(value), EnumMeta):
            return cls._encode_enum(value)
        if isinstance(value, dt.datetime):
            return cls._encode_datetime(name, value)
        if isinstance(value, dt.date):
            return cls._encode_date(value)
        if isinstance(value, set):
            return cls._encode_set(value)

        msg = f"Cannot encode {type(value)} value in or within field {name}: {value!r}"
        raise TypeError(msg)

    @staticmethod
    def _encode_enum(value: Enum) -> str:
        """Enums are encoded as a string containing their name."""
        return value.name

    @staticmethod
    def _encode_datetime(name, value: dt.datetime) -> str:
        """Timestamps are encoded in ISO format, with a 'Z' timezone specifier. Naive timestamps aren't allowed."""
        # Only allow tz-aware timestamps.
        if value.tzinfo is None:
            # Name refers to the outermost field, not necessarily a field on a (nested) dataclass.
            msg = f"Timestamp without timezone not supported in or within field {name}: {value}"
            raise ValueError(msg)
        # Always store with 'Z'.
        ts_utc = value.astimezone(dt.timezone.utc)
        return ts_utc.isoformat().replace("+00:00", "Z")

    @staticmethod
    def _encode_date(value: dt.date) -> str:
        """Dates are encoded in ISO format."""
        return value.isoformat()

    @staticmethod
    def _encode_set(value: set[T]) -> list[T]:
        """Sets are encoded as an array of the elements."""
        return list(value)

    def _encode_field_value(self, name: str, value: Any) -> str | None:
        if value is None:
            return None
        value_type = self._field_names_with_types[name]
        if value_type in (str, (str | None)):
            if isinstance(value, str):
                return value
            msg = f"Invalid value for field {name}, not a string: {value!r}"
            raise ValueError(msg)
        encoded_value = json.dumps(
            value,
            allow_nan=False,
            separators=(",", ":"),  # More compact than default, which includes a space after the colon (:).
            default=lambda o: self._encode_non_serializable(name, o),
        )
        # Handle encoding substituted values that encode as just a string (eg. timestamps); we just return the string.
        return json.loads(encoded_value) if encoded_value.startswith('"') else encoded_value

    def _object_data_and_failures(self, record: Record) -> tuple[dict[str, str], list[str]]:
        record_values = self._as_dict(record)
        encoded_fields = {
            field: self._encode_field_value(field, record_values[field]) for field in self._field_names_with_types
        }
        # We must return a value: strings are mandatory (not optional) as the type. As such, optional fields need to be
        # omitted from the data map if the value is None.
        data = {k: v for k, v in encoded_fields.items() if v is not None}
        if self._has_failures == list[str]:
            failures = record_values["failures"]
        elif self._has_failures == str:
            raw_failures = record_values["failures"]
            try:
                failures = json.loads(raw_failures) if raw_failures else []
            except json.decoder.JSONDecodeError:
                failures = [raw_failures]
        else:
            failures = []

        return data, failures

    def to_historical(self, record: Record) -> Historical:
        data, failures = self._object_data_and_failures(record)
        return Historical(
            workspace_id=self._workspace_id,
            job_run_id=self._job_run_id,
            object_type=self._object_type,
            object_id=self._object_id(record),
            data=data,
            failures=failures,
            owner=self._ownership.owner_of(record),
        )


class ProgressEncoder(Generic[Record]):
    def __init__(
        self,
        sql_backend: SqlBackend,
        ownership: Ownership[Record],
        klass: type[Record],
        run_id: int,
        workspace_id: int,
        catalog: str,
        schema: str = "multiworkspace",
        table: str = "historical",
    ) -> None:
        self._sql_backend = sql_backend
        self._klass = klass
        self._catalog = catalog
        self._schema = schema
        self._table = table
        self._encoder = HistoricalEncoder(
            job_run_id=run_id,
            workspace_id=workspace_id,
            ownership=ownership,
            klass=klass,
        )

    @property
    def full_name(self) -> str:
        return f"{self._catalog}.{self._schema}.{self._table}"

    def append_inventory_snapshot(self, snapshot: Iterable[Record]) -> None:
        history_records = [self._encode_record_as_historical(record) for record in snapshot]
        logger.debug(f"Appending {len(history_records)} {self._klass} record(s) to history.")
        # The mode is 'append'. This is documented as conflict-free.
        self._sql_backend.save_table(escape_sql_identifier(self.full_name), history_records, Historical, mode="append")

    def _encode_record_as_historical(self, record: Record) -> Historical:
        """Encode a snapshot record as a historical log entry."""
        return self._encoder.to_historical(record)



from __future__ import annotations

import datetime as dt
import logging
import os.path
import re
import sys
import webbrowser
from collections.abc import Iterator
from dataclasses import replace
from datetime import datetime, timedelta, timezone
from io import StringIO
from pathlib import Path
from typing import Any

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.parallel import ManyError
from databricks.labs.blueprint.wheels import ProductInfo, WheelsV2
from databricks.labs.lsql.dashboards import Dashboards
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import (
    Aborted,
    AlreadyExists,
    BadRequest,
    Cancelled,
    DataLoss,
    DeadlineExceeded,
    InternalError,
    InvalidParameterValue,
    NotFound,
    OperationFailed,
    PermissionDenied,
    RequestLimitExceeded,
    ResourceAlreadyExists,
    ResourceConflict,
    ResourceDoesNotExist,
    ResourceExhausted,
    TemporarilyUnavailable,
    TooManyRequests,
    Unauthenticated,
    Unknown,
)
from databricks.sdk.retries import retried
from databricks.sdk.service import compute, jobs
from databricks.sdk.service.jobs import Run, RunLifeCycleState, RunResultState
from databricks.sdk.service.workspace import ObjectType

import databricks
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.framework.tasks import Task
from databricks.labs.ucx.installer.logs import PartialLogRecord, parse_logs
from databricks.labs.ucx.installer.mixins import InstallationMixin

logger = logging.getLogger(__name__)

TEST_RESOURCE_PURGE_TIMEOUT = timedelta(hours=1)
TEST_NIGHTLY_CI_RESOURCES_PURGE_TIMEOUT = timedelta(hours=3)  # Buffer for debugging nightly integration test runs
# See https://docs.databricks.com/en/jobs/parameter-value-references.html#supported-value-references
EXTRA_TASK_PARAMS = {
    "job_id": "{{job_id}}",
    "run_id": "{{run_id}}",
    "start_time": "{{job.start_time.iso_datetime}}",
    "attempt": "{{job.repair_count}}",
    "parent_run_id": "{{parent_run_id}}",
}
DEBUG_NOTEBOOK = """# Databricks notebook source
# MAGIC %md
# MAGIC # Debug companion for UCX installation (see [README]({readme_link}))
# MAGIC
# MAGIC Production runs are supposed to be triggered through the following jobs: {job_links}
# MAGIC
# MAGIC **This notebook is overwritten with each UCX update/(re)install.**

# COMMAND ----------

# MAGIC %pip install {remote_wheel}
dbutils.library.restartPython()

# COMMAND ----------

import logging
from pathlib import Path
from databricks.sdk.config import with_user_agent_extra
from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.logger import install_logger
from databricks.labs.ucx.__about__ import __version__
from databricks.labs.ucx.contexts.workflow_task import RuntimeContext

install_logger()
with_user_agent_extra("cmd", "debug-notebook")
logging.getLogger("databricks").setLevel("DEBUG")

# ctx.<TAB> to see all available objects for you to
named_parameters = dict(config="/Workspace{config_file}")
ctx = RuntimeContext(named_parameters)

print(__version__)
"""

TEST_RUNNER_NOTEBOOK = """# Databricks notebook source
# MAGIC %pip install {remote_wheel}
dbutils.library.restartPython()

# COMMAND ----------

from databricks.labs.ucx.runtime import main

main(f'--config=/Workspace{config_file}',
     f'--workflow=' + dbutils.widgets.get('workflow'),
     f'--task=' + dbutils.widgets.get('task'),
     f'--job_id=' + dbutils.widgets.get('job_id'),
     f'--run_id=' + dbutils.widgets.get('run_id'),
     f'--start_time=' + dbutils.widgets.get('start_time'),
     f'--attempt=' + dbutils.widgets.get('attempt'),
     f'--parent_run_id=' + dbutils.widgets.get('parent_run_id'))
"""

EXPORT_TO_EXCEL_NOTEBOOK = """# Databricks notebook source
# MAGIC %md
# MAGIC ##### Exporter of UCX assessment results
# MAGIC ##### Instructions:
# MAGIC 1. Execute using an all-purpose cluster with Databricks Runtime 14 or higher.
# MAGIC 1. Hit **Run all** button and wait for completion.
# MAGIC 1. Go to the bottom of the notebook and click the Download UCX Results button.
# MAGIC
# MAGIC ##### Important:
# MAGIC Please note that this is only meant to serve as example code.
# MAGIC
# MAGIC Example code developed by **Databricks Shared Technical Services team**.

# COMMAND ----------

# DBTITLE 1,Installing Packages
# MAGIC %pip install {remote_wheel} -qqq
# MAGIC %pip install xlsxwriter -qqq
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

# DBTITLE 1,Libraries Import and Setting UCX
import os
import logging
import threading
import shutil
from pathlib import Path
from threading import Lock
from functools import partial

import pandas as pd
import xlsxwriter

from databricks.sdk.config import with_user_agent_extra
from databricks.labs.blueprint.logger import install_logger
from databricks.labs.blueprint.parallel import Threads
from databricks.labs.lsql.dashboards import Dashboards
from databricks.labs.lsql.lakeview.model import Dataset
from databricks.labs.ucx.contexts.workflow_task import RuntimeContext

# ctx
install_logger()
with_user_agent_extra("cmd", "export-assessment")
named_parameters = dict(config="/Workspace{config_file}")
ctx = RuntimeContext(named_parameters)
lock = Lock()

# COMMAND ----------

# DBTITLE 1,Assessment Export
FILE_NAME = "ucx_assessment_main.xlsx"
TMP_PATH = f"/Workspace{{ctx.installation.install_folder()}}/tmp/"
DOWNLOAD_PATH = "/dbfs/FileStore/excel-export"


def _cleanup() -> None:
    '''Move the temporary results file to the download path and clean up the temp directory.'''
    shutil.move(
        os.path.join(TMP_PATH, FILE_NAME),
        os.path.join(DOWNLOAD_PATH, FILE_NAME),
    )
    shutil.rmtree(TMP_PATH)


def _prepare_directories() -> None:
    '''Ensure that the necessary directories exist.'''
    os.makedirs(TMP_PATH, exist_ok=True)
    os.makedirs(DOWNLOAD_PATH, exist_ok=True)


def _to_excel(dataset: Dataset, writer: ...) -> None:
    '''Execute a SQL query and write the result to an Excel sheet.'''
    worksheet_name = dataset.display_name[:31]
    df = spark.sql(dataset.query).toPandas()
    with lock:
        df.to_excel(writer, sheet_name=worksheet_name, index=False)


def _render_export() -> None:
    '''Render an HTML link for downloading the results.'''
    html_content = '''
    <style>@font-face{{font-family:'DM Sans';src:url(https://cdn.bfldr.com/9AYANS2F/at/p9qfs3vgsvnp5c7txz583vgs/dm-sans-regular.ttf?auto=webp&format=ttf) format('truetype');font-weight:400;font-style:normal}}body{{font-family:'DM Sans',Arial,sans-serif}}.export-container{{text-align:center;margin-top:20px}}.export-container h2{{color:#1B3139;font-size:24px;margin-bottom:20px}}.export-container a{{display:inline-block;padding:12px 25px;background-color:#1B3139;color:#fff;text-decoration:none;border-radius:4px;font-size:18px;font-weight:500;transition:background-color 0.3s ease,transform:translateY(-2px) ease}}.export-container a:hover{{background-color:#FF3621;transform:translateY(-2px)}}</style>
    <div class="export-container"><h2>Export Results</h2><a href='{workspace_host}/files/excel-export/ucx_assessment_main.xlsx?o={workspace_id}' target='_blank' download>Download Results</a></div>

    '''
    displayHTML(html_content)


def export_results() -> None:
    '''Main method to export results to an Excel file.'''
    _prepare_directories()

    dashboard_path = (
        Path(ctx.installation.install_folder())
        / "dashboards/[UCX] UCX  Assessment (Main).lvdash.json"
    )
    dashboard = Dashboards(ctx.workspace_client)
    dashboard_datasets = dashboard.get_dashboard(dashboard_path).datasets
    try:
        target = TMP_PATH + "/ucx_assessment_main.xlsx"
        with pd.ExcelWriter(target, engine="xlsxwriter") as writer:
            tasks = []
            for dataset in dashboard_datasets:
                tasks.append(partial(_to_excel, dataset, writer))
                Threads.strict("exporting", tasks)
        _cleanup()
        _render_export()
    except Exception as e:
        print(f"Error exporting results ", e)

# COMMAND ----------

# DBTITLE 1,Data Export
export_results()
"""


class DeployedWorkflows:
    def __init__(self, ws: WorkspaceClient, install_state: InstallState):
        self._ws = ws
        self._install_state = install_state

    def run_workflow(self, step: str, skip_job_wait: bool = False, max_wait: timedelta = timedelta(minutes=20)) -> int:
        # this dunder variable is hiding this method from tracebacks, making it cleaner
        # for the user to see the actual error without too much noise.
        __tracebackhide__ = True  # pylint: disable=unused-variable
        job_id = int(self._install_state.jobs[step])
        logger.debug(f"starting {step} job: {self._ws.config.host}#job/{job_id}")
        job_initial_run = self._ws.jobs.run_now(job_id)
        run_id = job_initial_run.run_id
        if not run_id:
            raise NotFound(f"job run not found for {step}")
        run_url = f"{self._ws.config.host}#job/{job_id}/runs/{run_id}"
        logger.info(f"Started {step} job: {run_url}")
        if skip_job_wait:
            return run_id
        try:
            logger.debug(f"Waiting for completion of {step} job: {run_url}")
            job_run = self._ws.jobs.wait_get_run_job_terminated_or_skipped(run_id=run_id, timeout=max_wait)
            self._log_completed_job(step, run_id, job_run)
            return run_id
        except TimeoutError:
            logger.warning(f"Timeout while waiting for {step} job to complete: {run_url}")
            logger.info('---------- REMOTE LOGS --------------')
            self._relay_logs(step, run_id)
            logger.info('------ END REMOTE LOGS (SO FAR) -----')
            raise
        except OperationFailed as err:
            logger.info('---------- REMOTE LOGS --------------')
            self._relay_logs(step, run_id)
            logger.info('---------- END REMOTE LOGS ----------')
            job_run = self._ws.jobs.get_run(run_id)
            raise self._infer_error_from_job_run(job_run) from err

    @staticmethod
    def _log_completed_job(step: str, run_id: int, job_run: Run) -> None:
        if job_run.state:
            result_state = job_run.state.result_state or "N/A"
            state_message = job_run.state.state_message
            state_description = f"{result_state} ({state_message})" if state_message else f"{result_state}"
            logger.info(f"Completed {step} job run {run_id} with state: {state_description}")
        else:
            logger.warning(f"Completed {step} job run {run_id} but end state is unknown.")
        if job_run.start_time or job_run.end_time:
            start_time = (
                datetime.fromtimestamp(job_run.start_time / 1000, tz=timezone.utc) if job_run.start_time else None
            )
            end_time = datetime.fromtimestamp(job_run.end_time / 1000, tz=timezone.utc) if job_run.end_time else None
            if job_run.run_duration:
                duration = timedelta(milliseconds=job_run.run_duration)
            elif start_time and end_time:
                duration = end_time - start_time
            else:
                duration = None
            logger.info(
                f"Completed {step} job run {run_id} duration: {duration or 'N/A'} ({start_time or 'N/A'} thru {end_time or 'N/A'})"
            )

    def repair_run(self, workflow, verify_timeout: timedelta = timedelta(minutes=2)):
        try:
            job_id, run_id = self._repair_workflow(workflow, verify_timeout)
            run_details = self._ws.jobs.get_run(run_id=run_id, include_history=True)
            self._handle_repair_run(run_details, job_id, run_id, workflow)
        except InvalidParameterValue as e:
            logger.warning(f"Skipping {workflow}: {e}")
        except TimeoutError:
            logger.warning(f"Skipping the {workflow} due to time out. Please try after sometime")

    def _handle_repair_run(self, run_details, job_id, run_id, workflow):
        if run_details.repair_history:
            latest_repair_run_id = run_details.repair_history[-1].id
            job_url = f"{self._ws.config.host}#job/{job_id}/run/{run_id}"
            logger.debug(f"Repairing {workflow} job: {job_url}")
            self._ws.jobs.repair_run(run_id=run_id, rerun_all_failed_tasks=True, latest_repair_id=latest_repair_run_id)
            webbrowser.open(job_url)
        else:
            logger.warning(f"No repair history found for run_id={run_id}")

    def latest_job_status(self) -> list[dict]:
        latest_status = []
        for step, job_id in self._install_state.jobs.items():
            job_state = None
            start_time = None
            try:
                job_runs = list(self._ws.jobs.list_runs(job_id=int(job_id), limit=1))
            except InvalidParameterValue as e:
                logger.warning(f"skipping {step}: {e}")
                continue
            if job_runs:
                state = job_runs[0].state
                if state and state.result_state:
                    job_state = state.result_state.name
                elif state and state.life_cycle_state:
                    job_state = state.life_cycle_state.name
                if job_runs[0].start_time:
                    start_time = job_runs[0].start_time / 1000
            latest_status.append(
                {
                    "step": step,
                    "state": "UNKNOWN" if not (job_runs and job_state) else job_state,
                    "started": (
                        "<never run>" if not (job_runs and start_time) else self._readable_timedelta(start_time)
                    ),
                }
            )
        return latest_status

    def validate_step(self, step: str, *, timeout: dt.timedelta = dt.timedelta(minutes=20)) -> bool:
        """Validate a workflow has completed successfully.

        If none of the job runs belonging to the workflow did not finish successfully (yet) and at least one job run is
        running or pending, we wait for the given timeout for that job run to finish. Thereafter, if none of the running
        or pending job runs, finished within the timeout, we consider the step to be invalid, i.e. we return `False`.

        Args :
            step (str) : The workflow name; step in the migration process.
            timeout (datetime.timedelta, optional) : The timeout to wait for a running or pending job to finish.
                Defaults to 20 minutes.

        Returns :
            bool : True if step is validate. False otherwise.
        """
        job_id = int(self._install_state.jobs[step])
        logger.debug(f"Validating {step} workflow: {self._ws.config.host}#job/{job_id}")
        current_runs = list(self._ws.jobs.list_runs(completed_only=False, job_id=job_id))
        for run in current_runs:
            if run.state and run.state.result_state == RunResultState.SUCCESS:
                return True
        for run in current_runs:
            if (
                run.run_id
                and run.state
                and run.state.life_cycle_state in (RunLifeCycleState.RUNNING, RunLifeCycleState.PENDING)
            ):
                logger.info("Identified a run in progress waiting for run completion")
                try:
                    self._ws.jobs.wait_get_run_job_terminated_or_skipped(run_id=run.run_id, timeout=timeout)
                except TimeoutError:
                    return False
                run_new_state = self._ws.jobs.get_run(run_id=run.run_id).state
                return run_new_state is not None and run_new_state.result_state == RunResultState.SUCCESS
        return False

    def relay_logs(self, workflow: str | None = None):
        latest_run = None
        if not workflow:
            runs = []
            for step in self._install_state.jobs:
                try:
                    _, latest_run = self._latest_job_run(step)
                    runs.append((step, latest_run))
                except InvalidParameterValue:
                    continue
            if not runs:
                logger.warning("No jobs to relay logs for")
                return
            runs = sorted(runs, key=lambda x: x[1].start_time, reverse=True)
            workflow, latest_run = runs[0]
        if not latest_run:
            assert workflow is not None
            _, latest_run = self._latest_job_run(workflow)
        self._relay_logs(workflow, latest_run.run_id)

    def _relay_logs(self, workflow, run_id):
        for record in self._fetch_last_run_attempt_logs(workflow, run_id):
            task_logger = logging.getLogger(record.component)
            MaxedStreamHandler.install_handler(task_logger)
            task_logger.setLevel(logger.getEffectiveLevel())
            log_level = logging.getLevelName(record.level)
            task_logger.log(log_level, record.message)
        MaxedStreamHandler.uninstall_handlers()

    def _fetch_last_run_attempt_logs(self, workflow: str, run_id: str) -> Iterator[PartialLogRecord]:
        """Fetch the logs for the last run attempt."""
        run_folders = self._get_log_run_folders(workflow, run_id)
        if not run_folders:
            return
        # sort folders based on the last repair attempt
        last_attempt = sorted(run_folders, key=lambda _: int(_.split('-')[-1]), reverse=True)[0]
        for object_info in self._ws.workspace.list(last_attempt):
            if not object_info.path:
                continue
            if '.log' not in object_info.path:
                continue
            task_name = os.path.basename(object_info.path).split('.')[0]
            with self._ws.workspace.download(object_info.path) as raw_file:
                text_io = StringIO(raw_file.read().decode())
            for record in parse_logs(text_io):
                yield replace(record, component=f'{record.component}:{task_name}')

    def _get_log_run_folders(self, workflow: str, run_id: str) -> list[str]:
        """Get the log run folders.

        The log run folders are located in the installation folder under the logs directory. Each workflow has a log run
        folder for each run id. Multiple runs occur for repair runs.
        """
        log_path = f"{self._install_state.install_folder()}/logs/{workflow}"
        try:
            # Ensure any exception is triggered early.
            log_path_objects = list(self._ws.workspace.list(log_path))
        except ResourceDoesNotExist:
            logger.warning(f"Cannot fetch logs as folder {log_path} does not exist")
            return []
        run_folders = []
        for run_folder in log_path_objects:
            if not run_folder.path or run_folder.object_type != ObjectType.DIRECTORY:
                continue
            if f"run-{run_id}-" not in run_folder.path:
                continue
            run_folders.append(run_folder.path)
        return run_folders

    @staticmethod
    def _readable_timedelta(epoch):
        when = datetime.utcfromtimestamp(epoch)
        duration = datetime.now() - when
        data = {}
        data["days"], remaining = divmod(duration.total_seconds(), 86_400)
        data["hours"], remaining = divmod(remaining, 3_600)
        data["minutes"], data["seconds"] = divmod(remaining, 60)

        time_parts = ((name, round(value)) for (name, value) in data.items())
        time_parts = [f"{value} {name[:-1] if value == 1 else name}" for name, value in time_parts if value > 0]
        if len(time_parts) > 0:
            time_parts.append("ago")
        if time_parts:
            return " ".join(time_parts)
        return "less than 1 second ago"

    def _repair_workflow(self, workflow, verify_timeout):
        job_id, latest_job_run = self._latest_job_run(workflow)
        retry_on_attribute_error = retried(on=[AttributeError], timeout=verify_timeout)
        retried_check = retry_on_attribute_error(self._get_result_state)
        state_value = retried_check(job_id)
        logger.info(f"The status for the latest run is {state_value}")
        if state_value != "FAILED":
            raise InvalidParameterValue("job is not in FAILED state hence skipping repair")
        run_id = latest_job_run.run_id
        return job_id, run_id

    def _latest_job_run(self, workflow: str):
        job_id = self._install_state.jobs.get(workflow)
        if not job_id:
            raise InvalidParameterValue("job does not exists hence skipping repair")
        job_runs = list(self._ws.jobs.list_runs(job_id=job_id, limit=1))
        if not job_runs:
            raise InvalidParameterValue("job is not initialized yet. Can't trigger repair run now")
        latest_job_run = job_runs[0]
        return job_id, latest_job_run

    def _get_result_state(self, job_id):
        job_runs = list(self._ws.jobs.list_runs(job_id=job_id, limit=1))
        latest_job_run = job_runs[0]
        if not latest_job_run.state.result_state:
            raise AttributeError("no result state in job run")
        job_state = latest_job_run.state.result_state.value
        return job_state

    def _infer_error_from_job_run(self, job_run) -> Exception:
        errors: list[Exception] = []
        timeouts: list[DeadlineExceeded] = []
        assert job_run.tasks is not None
        for run_task in job_run.tasks:
            error = self._infer_error_from_task_run(run_task)
            if not error:
                continue
            if isinstance(error, DeadlineExceeded):
                timeouts.append(error)
                continue
            errors.append(error)
        assert job_run.state is not None
        assert job_run.state.state_message is not None
        if len(errors) == 1:
            return errors[0]
        all_errors = errors + timeouts
        if len(all_errors) == 0:
            return Unknown(job_run.state.state_message)
        return ManyError(all_errors)

    def _infer_error_from_task_run(self, run_task: jobs.RunTask) -> Exception | None:
        if not run_task.state:
            return None
        if run_task.state.result_state == jobs.RunResultState.TIMEDOUT:
            msg = f"{run_task.task_key}: The run was stopped after reaching the timeout"
            return DeadlineExceeded(msg)
        if run_task.state.result_state != jobs.RunResultState.FAILED:
            return None
        assert run_task.run_id is not None
        run_output = self._ws.jobs.get_run_output(run_task.run_id)
        if not run_output:
            msg = f'No run output. {run_task.state.state_message}'
            return InternalError(msg)
        if logger.isEnabledFor(logging.DEBUG):
            if run_output.error_trace:
                sys.stderr.write(run_output.error_trace)
        if not run_output.error:
            msg = f'No error in run output. {run_task.state.state_message}'
            return InternalError(msg)
        return self._infer_task_exception(f"{run_task.task_key}: {run_output.error}")

    @staticmethod
    def _infer_task_exception(haystack: str) -> Exception:
        needles: list[type[Exception]] = [
            BadRequest,
            Unauthenticated,
            PermissionDenied,
            NotFound,
            ResourceConflict,
            TooManyRequests,
            Cancelled,
            databricks.sdk.errors.NotImplemented,
            InternalError,
            TemporarilyUnavailable,
            DeadlineExceeded,
            InvalidParameterValue,
            ResourceDoesNotExist,
            Aborted,
            AlreadyExists,
            ResourceAlreadyExists,
            ResourceExhausted,
            RequestLimitExceeded,
            Unknown,
            DataLoss,
            ValueError,
            KeyError,
        ]
        constructors: dict[re.Pattern, type[Exception]] = {
            re.compile(r".*\[TABLE_OR_VIEW_NOT_FOUND] (.*)"): NotFound,
            re.compile(r".*\[SCHEMA_NOT_FOUND] (.*)"): NotFound,
            re.compile(r".*\[TimeoutException] (.*)"): TimeoutError,
        }
        for klass in needles:
            constructors[re.compile(f".*{klass.__name__}: (.*)")] = klass
        for pattern, klass in constructors.items():
            match = pattern.match(haystack)
            if match:
                return klass(match.group(1))
        return Unknown(haystack)


class WorkflowsDeployment(InstallationMixin):
    def __init__(
        self,
        config: WorkspaceConfig,
        installation: Installation,
        install_state: InstallState,
        ws: WorkspaceClient,
        wheels: WheelsV2,
        product_info: ProductInfo,
        tasks: list[Task],
    ):
        self._config = config
        self._installation = installation
        self._ws = ws
        self._install_state = install_state
        self._wheels = wheels
        self._product_info = product_info
        self._tasks = tasks
        self._this_file = Path(__file__)
        super().__init__(config, installation, ws)

    def create_jobs(self) -> None:
        remote_wheels = self._upload_wheel()
        desired_workflows = {t.workflow for t in self._tasks if t.cloud_compatible(self._ws.config)}

        wheel_runner = ""
        if self._config.override_clusters:
            wheel_runner = self._upload_wheel_runner(remote_wheels)
        for workflow_name in desired_workflows:
            settings = self._job_settings(workflow_name, remote_wheels)
            if self._config.override_clusters:
                settings = self._apply_cluster_overrides(
                    workflow_name,
                    settings,
                    self._config.override_clusters,
                    wheel_runner,
                )
            self._deploy_workflow(workflow_name, settings)

        self.remove_jobs(keep=desired_workflows)
        self._install_state.save()
        self._create_debug(remote_wheels)
        self._create_export(remote_wheels)
        self._create_readme()

    def remove_jobs(self, *, keep: set[str] | None = None) -> None:
        for workflow_name, job_id in self._install_state.jobs.items():
            if keep and workflow_name in keep:
                continue
            try:
                if not self._is_managed_job_failsafe(int(job_id)):
                    logger.warning(f"Corrupt installation state. Skipping job_id={job_id} as it is not managed by UCX")
                    continue
                logger.info(f"Removing job_id={job_id}, as it is no longer needed")
                self._ws.jobs.delete(job_id)
            except InvalidParameterValue:
                logger.warning(f"step={workflow_name} does not exist anymore for some reason")
                continue

    # see https://github.com/databrickslabs/ucx/issues/2667
    def _is_managed_job_failsafe(self, job_id: int) -> bool:
        install_folder = self._installation.install_folder()
        try:
            return self._is_managed_job(job_id, install_folder)
        except ResourceDoesNotExist:
            return False
        except InvalidParameterValue:
            return False

    def _is_managed_job(self, job_id: int, install_folder: str) -> bool:
        job = self._ws.jobs.get(job_id)
        if not job.settings or not job.settings.tasks:
            return False
        for task in job.settings.tasks:
            if task.notebook_task and task.notebook_task.notebook_path.startswith(install_folder):
                return True
            if task.python_wheel_task and task.python_wheel_task.package_name == "databricks_labs_ucx":
                return True
        return False

    @property
    def _config_file(self):
        return f"{self._installation.install_folder()}/config.yml"

    def _is_testing(self):
        return self._product_info.product_name() != "ucx"

    @staticmethod
    def _is_nightly():
        ci_env = os.getenv("TEST_NIGHTLY")
        return ci_env is not None and ci_env.lower() == "true"

    @classmethod
    def _get_test_purge_time(cls) -> str:
        # Duplicate of mixins.fixtures.get_test_purge_time(); we don't want to import pytest as a transitive dependency.
        timeout = TEST_NIGHTLY_CI_RESOURCES_PURGE_TIMEOUT if cls._is_nightly() else TEST_RESOURCE_PURGE_TIMEOUT
        now = datetime.now(timezone.utc)
        purge_deadline = now + timeout
        # Round UP to the next hour boundary: that is when resources will be deleted.
        purge_hour = purge_deadline + (datetime.min.replace(tzinfo=timezone.utc) - purge_deadline) % timedelta(hours=1)
        return purge_hour.strftime("%Y%m%d%H")

    def _create_readme(self) -> None:
        debug_notebook_link = self._installation.workspace_markdown_link('debug notebook', 'DEBUG.py')
        markdown = [
            "# UCX - The Unity Catalog Migration Assistant",
            f'To troubleshoot, see {debug_notebook_link}.\n',
            "Here are the URLs and descriptions of workflows that trigger various stages of migration.",
            "All jobs are defined with necessary cluster configurations and DBR versions.\n",
        ]
        for step_name in self._step_list():
            if step_name not in self._install_state.jobs:
                logger.warning(f"Skipping step '{step_name}' since it was not deployed.")
                continue
            job_id = self._install_state.jobs[step_name]
            dashboard_link = ""
            dashboard_link = self._create_dashboard_links(step_name, dashboard_link)
            job_link = f"[{self._name(step_name)}]({self._ws.config.host}#job/{job_id})"
            markdown.append("---\n\n")
            markdown.append(f"## {job_link}\n\n")
            markdown.append(f"{dashboard_link}")
            markdown.append("\nThe workflow consists of the following separate tasks:\n\n")
            for task in self._tasks:
                if self._is_testing() and task.is_testing():
                    continue
                if task.workflow != step_name:
                    continue
                doc = self._config.replace_inventory_variable(task.doc)
                markdown.append(f"### `{task.name}`\n\n")
                markdown.append(f"{doc}\n")
                markdown.append("\n\n")
        preamble = ["# Databricks notebook source", "# MAGIC %md"]
        intro = "\n".join(preamble + [f"# MAGIC {line}" for line in markdown])
        self._installation.upload('README.py', intro.encode('utf8'))

    def _create_dashboard_links(self, step_name, dashboard_link):
        dashboards_per_step = [d for d in self._install_state.dashboards.keys() if d.startswith(step_name)]
        for dash in dashboards_per_step:
            if len(dashboard_link) == 0:
                dashboard_link += "Go to the one of the following dashboards after running the job:\n"
            first, second = dash.replace("_", " ").title().split()
            dashboard_url = Dashboards(self._ws).get_url(self._install_state.dashboards[dash])
            dashboard_link += f"  - [{first} ({second}) dashboard]({dashboard_url})\n"
        return dashboard_link

    def _step_list(self) -> list[str]:
        step_list = []
        for task in self._tasks:
            if self._is_testing() and task.is_testing():
                continue
            if task.workflow not in step_list:
                step_list.append(task.workflow)
        return step_list

    def _job_cluster_spark_conf(self, cluster_key: str):
        conf_from_installation = self._config.spark_conf if self._config.spark_conf else {}
        if cluster_key == "main":
            spark_conf = {
                "spark.databricks.cluster.profile": "singleNode",
                "spark.master": "local[*]",
            }
            return spark_conf | conf_from_installation
        if cluster_key == "tacl":
            return {"spark.databricks.acl.sqlOnly": "true"} | conf_from_installation
        if cluster_key == "user_isolation":
            return {"spark.sql.sources.parallelPartitionDiscovery.parallelism": "200"} | conf_from_installation
        return conf_from_installation

    # Workflow creation might fail on an InternalError with no message
    @retried(on=[InternalError], timeout=timedelta(minutes=2))
    def _deploy_workflow(self, step_name: str, settings):
        if step_name in self._install_state.jobs:
            try:
                job_id = int(self._install_state.jobs[step_name])
                logger.info(f"Updating configuration for step={step_name} job_id={job_id}")
                return self._ws.jobs.reset(job_id, jobs.JobSettings(**settings))
            except InvalidParameterValue:
                del self._install_state.jobs[step_name]
                logger.warning(f"step={step_name} does not exist anymore for some reason")
                return self._deploy_workflow(step_name, settings)
        logger.info(f"Creating new job configuration for step={step_name}")
        new_job = self._ws.jobs.create(**settings)
        assert new_job.job_id is not None
        self._install_state.jobs[step_name] = str(new_job.job_id)
        return None

    @staticmethod
    def _library_dep_order(library: str):
        match library:
            case library if 'sdk' in library:
                return 0
            case library if 'blueprint' in library:
                return 1
            case library if 'sqlglot' in library:
                return 2
            case _:
                return 3

    def _upload_wheel(self):
        wheel_paths = []
        with self._wheels:
            if self._config.upload_dependencies:
                wheel_paths = self._wheels.upload_wheel_dependencies(["databricks", "sqlglot", "astroid"])
            wheel_paths.sort(key=WorkflowsDeployment._library_dep_order)
            wheel_paths.append(self._wheels.upload_to_wsfs())
            wheel_paths = [f"/Workspace{wheel}" for wheel in wheel_paths]
            return wheel_paths

    def _upload_wheel_runner(self, remote_wheels: list[str]) -> str:
        # TODO: we have to be doing this workaround until ES-897453 is solved in the platform
        remote_wheels_str = " ".join(remote_wheels)
        code = TEST_RUNNER_NOTEBOOK.format(remote_wheel=remote_wheels_str, config_file=self._config_file).encode("utf8")
        return self._installation.upload(f"wheels/wheel-test-runner-{self._product_info.version()}.py", code)

    @staticmethod
    def _apply_cluster_overrides(
        workflow_name: str,
        settings: dict[str, Any],
        overrides: dict[str, str],
        wheel_runner: str,
    ) -> dict:
        settings["job_clusters"] = [_ for _ in settings["job_clusters"] if _.job_cluster_key not in overrides]
        for job_task in settings["tasks"]:
            if job_task.job_cluster_key is None:
                continue
            if job_task.job_cluster_key in overrides:
                job_task.existing_cluster_id = overrides[job_task.job_cluster_key]
                job_task.job_cluster_key = None
                job_task.libraries = None
            if job_task.python_wheel_task is not None:
                job_task.python_wheel_task = None
                widget_values = {"task": job_task.task_key, 'workflow': workflow_name} | EXTRA_TASK_PARAMS
                job_task.notebook_task = jobs.NotebookTask(notebook_path=wheel_runner, base_parameters=widget_values)
        return settings

    def _job_settings(self, step_name: str, remote_wheels: list[str]) -> dict[str, Any]:
        email_notifications = None
        if not self._config.override_clusters and "@" in self._my_username:
            # set email notifications only if we're running the real
            # installation and not the integration test.
            email_notifications = jobs.JobEmailNotifications(
                on_success=[self._my_username], on_failure=[self._my_username]
            )

        job_tasks = []
        job_clusters: set[str] = {Task.job_cluster}
        for task in self._tasks:
            if task.workflow != step_name:
                continue
            job_clusters.add(task.job_cluster)
            job_tasks.append(self._job_task(task, remote_wheels))
        job_tasks.append(self._job_parse_logs_task(job_tasks, step_name, remote_wheels))
        version = self._product_info.version()
        version = version if not self._ws.config.is_gcp else version.replace("+", "-")
        tags = {"version": f"v{version}"}
        if self._is_testing():
            # add RemoveAfter tag for test job cleanup
            date_to_remove = self._get_test_purge_time()
            tags.update({"RemoveAfter": date_to_remove})
        return {
            "name": self._name(step_name),
            "tags": tags,
            "job_clusters": self._job_clusters(job_clusters),
            "email_notifications": email_notifications,
            "tasks": job_tasks,
        }

    def _job_task(self, task: Task, remote_wheels: list[str]) -> jobs.Task:
        jobs_task = jobs.Task(
            task_key=task.name,
            job_cluster_key=task.job_cluster,
            depends_on=[jobs.TaskDependency(task_key=d) for d in task.dependencies()],
        )
        if task.notebook:
            return self._job_notebook_task(jobs_task, task)
        return self._job_wheel_task(jobs_task, task.workflow, remote_wheels)

    def _job_notebook_task(self, jobs_task: jobs.Task, task: Task) -> jobs.Task:
        assert task.notebook is not None
        local_notebook = self._this_file.parent.parent / task.notebook
        with local_notebook.open("rb") as f:
            remote_notebook = self._installation.upload(local_notebook.name, f.read())
        return replace(
            jobs_task,
            notebook_task=jobs.NotebookTask(
                notebook_path=remote_notebook,
                # ES-872211: currently, we cannot read WSFS files from Scala context
                base_parameters={
                    "task": task.name,
                    "config": f"/Workspace{self._config_file}",
                }
                | EXTRA_TASK_PARAMS,
            ),
        )

    def _job_wheel_task(self, jobs_task: jobs.Task, workflow: str, remote_wheels: list[str]) -> jobs.Task:
        libraries = []
        for wheel in remote_wheels:
            libraries.append(compute.Library(whl=wheel))
        named_parameters = {
            "config": f"/Workspace{self._config_file}",
            "workflow": workflow,
            "task": jobs_task.task_key,
        }
        return replace(
            jobs_task,
            libraries=libraries,
            python_wheel_task=jobs.PythonWheelTask(
                package_name="databricks_labs_ucx",
                entry_point="runtime",  # [project.entry-points.databricks] in pyproject.toml
                named_parameters=named_parameters | EXTRA_TASK_PARAMS,
            ),
        )

    def _job_clusters(self, names: set[str]):
        clusters = []
        if "main" in names:
            clusters.append(
                jobs.JobCluster(
                    job_cluster_key="main",
                    new_cluster=compute.ClusterSpec(
                        data_security_mode=compute.DataSecurityMode.LEGACY_SINGLE_USER_STANDARD,
                        spark_conf=self._job_cluster_spark_conf("main"),
                        custom_tags={"ResourceClass": "SingleNode"},
                        num_workers=0,
                        policy_id=self._config.policy_id,
                    ),
                )
            )
        if "tacl" in names:
            clusters.append(
                jobs.JobCluster(
                    job_cluster_key="tacl",
                    new_cluster=compute.ClusterSpec(
                        data_security_mode=compute.DataSecurityMode.LEGACY_TABLE_ACL,
                        spark_conf=self._job_cluster_spark_conf("tacl"),
                        num_workers=1,  # ShowPermissionsCommand needs a worker
                        policy_id=self._config.policy_id,
                    ),
                )
            )
        if "user_isolation" in names:
            clusters.append(
                jobs.JobCluster(
                    job_cluster_key="user_isolation",
                    new_cluster=compute.ClusterSpec(
                        data_security_mode=compute.DataSecurityMode.USER_ISOLATION,
                        spark_conf=self._job_cluster_spark_conf("user_isolation"),
                        policy_id=self._config.policy_id,
                        autoscale=compute.AutoScale(
                            max_workers=self._config.max_workers,
                            min_workers=self._config.min_workers,
                        ),
                    ),
                )
            )
        return clusters

    def _job_parse_logs_task(self, job_tasks: list[jobs.Task], workflow: str, remote_wheels: list[str]) -> jobs.Task:
        jobs_task = jobs.Task(
            task_key="parse_logs",
            job_cluster_key=Task.job_cluster,
            # The task dependents on all previous tasks.
            depends_on=[jobs.TaskDependency(task_key=task.task_key) for task in job_tasks],
            run_if=jobs.RunIf.ALL_DONE,
        )
        return self._job_wheel_task(jobs_task, workflow, remote_wheels)

    def _create_debug(self, remote_wheels: list[str]):
        readme_link = self._installation.workspace_link('README')
        job_links = ", ".join(
            f"[{self._name(step_name)}]({self._ws.config.host}#job/{job_id})"
            for step_name, job_id in self._install_state.jobs.items()
        )
        remote_wheels_str = " ".join(remote_wheels)
        content = DEBUG_NOTEBOOK.format(
            remote_wheel=remote_wheels_str, readme_link=readme_link, job_links=job_links, config_file=self._config_file
        ).encode("utf8")
        self._installation.upload('DEBUG.py', content)

    def _create_export(self, remote_wheels: list[str]):
        remote_wheels_str = " ".join(remote_wheels)
        content = EXPORT_TO_EXCEL_NOTEBOOK.format(
            remote_wheel=remote_wheels_str,
            config_file=self._config_file,
            workspace_host=self._ws.config.host,
            workspace_id=self._ws.get_workspace_id(),
        ).encode("utf8")
        self._installation.upload('EXPORT_ASSESSMENT_TO_EXCEL.py', content)


class MaxedStreamHandler(logging.StreamHandler):

    MAX_STREAM_SIZE = 2**20 - 2**6  # 1 Mb minus some buffer
    _installed_handlers: dict[str, tuple[logging.Logger, MaxedStreamHandler]] = {}
    _sent_bytes = 0

    @classmethod
    def install_handler(cls, logger_: logging.Logger):
        if logger_.handlers:
            # already installed ?
            installed = next((h for h in logger_.handlers if isinstance(h, MaxedStreamHandler)), None)
            if installed:
                return
            # any handler to override ?
            handler = next((h for h in logger_.handlers if isinstance(h, logging.StreamHandler)), None)
            if handler:
                to_install = MaxedStreamHandler(handler)
                cls._installed_handlers[logger_.name] = (logger_, to_install)
                logger_.removeHandler(handler)
                logger_.addHandler(to_install)
                return
        if logger_.parent:
            cls.install_handler(logger_.parent)
        if logger_.root:
            cls.install_handler(logger_.root)

    @classmethod
    def uninstall_handlers(cls):
        for logger_, handler in cls._installed_handlers.values():
            logger_.removeHandler(handler)
            logger_.addHandler(handler.original_handler)
        cls._installed_handlers.clear()
        cls._sent_bytes = 0

    def __init__(self, original_handler: logging.StreamHandler):
        super().__init__()
        self._original_handler = original_handler

    @property
    def original_handler(self):
        return self._original_handler

    def emit(self, record):
        try:
            msg = self.format(record) + self.terminator
            if self._prevent_overflow(msg):
                return
            self.stream.write(msg)
            self.flush()
        except RecursionError:  # See issue 36272
            raise
        # the below is copied from Python source
        # so ensuring not to break the logging logic
        # pylint: disable=broad-exception-caught
        except Exception:
            self.handleError(record)

    def _prevent_overflow(self, msg: str):
        data = msg.encode("utf-8")
        if self._sent_bytes + len(data) > self.MAX_STREAM_SIZE:
            # ensure readers are aware of why the logs are incomplete
            self.stream.write(f"MAX LOGS SIZE REACHED: {self._sent_bytes} bytes!!!")
            self.flush()
            return True
        return False



import logging
from dataclasses import replace

from databricks.labs.blueprint.installation import Installation, SerdeError

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.sql import LegacyQuery, UpdateQueryRequestQuery
from databricks.sdk.errors.platform import DatabricksError

from databricks.labs.ucx.assessment.dashboards import Dashboard, RedashDashboardCrawler, Query
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.linters.from_table import FromTableSqlLinter

logger = logging.getLogger(__name__)


class Redash:
    MIGRATED_TAG = "Migrated by UCX"

    def __init__(
        self,
        index: TableMigrationIndex,
        ws: WorkspaceClient,
        installation: Installation,
        dashboard_crawler: RedashDashboardCrawler,
    ):
        self._index = index
        self._ws = ws
        self._installation = installation
        self._crawler = dashboard_crawler

    def migrate_dashboards(self, *dashboard_ids: str) -> None:
        for dashboard in self._list_dashboards(*dashboard_ids):
            if self.MIGRATED_TAG in dashboard.tags:
                logger.debug(f"Dashboard {dashboard.name} already migrated by UCX")
                continue
            for query in self._crawler.list_queries(dashboard):
                self._fix_query(query)
            self._ws.dashboards.update(dashboard.id, tags=self._get_migrated_tags(dashboard.tags))

    def revert_dashboards(self, *dashboard_ids: str) -> None:
        for dashboard in self._list_dashboards(*dashboard_ids):  # Refresh for up-to-date tags
            if self.MIGRATED_TAG not in dashboard.tags:
                logger.debug(f"Dashboard {dashboard.name} was not migrated by UCX")
                continue
            for query in self._crawler.list_queries(dashboard):
                self._revert_query(query)
            self._ws.dashboards.update(dashboard.id, tags=self._get_original_tags(dashboard.tags))

    def _list_dashboards(self, *dashboard_ids: str) -> list[Dashboard]:
        """List the Redash dashboards."""
        # Cached property is not used as this class in used from the CLI, thus called once per Python process
        dashboards = [d for d in self._crawler.snapshot() if not dashboard_ids or d.id in dashboard_ids]
        return dashboards

    def _fix_query(self, query: Query) -> None:
        assert query.id is not None
        assert query.query is not None
        # query already migrated
        if query.tags is not None and self.MIGRATED_TAG in query.tags:
            return
        # backup the query
        self._installation.save(query, filename=f'backup/queries/{query.id}.json')
        from_table = FromTableSqlLinter(self._index, self._get_session_state(query))
        new_query = UpdateQueryRequestQuery(
            query_text=from_table.apply(query.query),
            tags=self._get_migrated_tags(query.tags),
        )
        try:
            self._ws.queries.update(
                query.id,
                update_mask="query_text,tags",
                query=new_query,
            )
        except DatabricksError:
            logger.warning(f"Cannot upgrade {query.name}")
            return

    @staticmethod
    def _get_session_state(query: Query) -> CurrentSessionState:
        session_state = CurrentSessionState()
        if query.catalog:
            session_state = replace(session_state, catalog=query.catalog)
        if query.schema:
            session_state = replace(session_state, schema=query.schema)
        return session_state

    def _revert_query(self, query: Query) -> None:
        assert query.id is not None
        assert query.query is not None
        if self.MIGRATED_TAG not in (query.tags or []):
            logger.debug(f"Query {query.name} was not migrated by UCX")
            return
        backup_query: Query | LegacyQuery
        try:
            backup_query = self._installation.load(Query, filename=f'backup/queries/{query.id}.json')
        except SerdeError:
            # Previous versions store queries as LegacyQuery
            backup_query = self._installation.load(LegacyQuery, filename=f'backup/queries/{query.id}.json')
        update_query = UpdateQueryRequestQuery(query_text=backup_query.query, tags=self._get_original_tags(query.tags))
        try:
            self._ws.queries.update(query.id, update_mask="query_text,tags", query=update_query)
        except DatabricksError:
            logger.warning(f"Cannot restore {query.name} from backup")
            return

    def _get_migrated_tags(self, tags: list[str] | None) -> list[str]:
        out = [self.MIGRATED_TAG]
        if tags:
            out.extend(tags)
        return out

    def _get_original_tags(self, tags: list[str] | None) -> list[str] | None:
        if tags is None:
            return None
        return [tag for tag in tags if tag != self.MIGRATED_TAG]



import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from typing import TypeVar

from astroid import Attribute, Call, Const, Name, NodeNG  # type: ignore
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex, TableMigrationStatus
from databricks.labs.ucx.source_code.base import (
    Advice,
    Advisory,
    Deprecation,
    CurrentSessionState,
    SqlLinter,
    Fixer,
    UsedTable,
    UsedTableNode,
    TableSqlCollector,
    DfsaSqlCollector,
)
from databricks.labs.ucx.source_code.linters.directfs import DIRECT_FS_ACCESS_PATTERNS, DirectFsAccessNode
from databricks.labs.ucx.source_code.python.python_infer import InferredValue
from databricks.labs.ucx.source_code.linters.from_table import FromTableSqlLinter
from databricks.labs.ucx.source_code.python.python_ast import (
    Tree,
    TreeHelper,
    MatchingVisitor,
    PythonLinter,
    TablePyCollector,
    DfsaPyCollector,
)

logger = logging.getLogger(__name__)


@dataclass
class _TableNameMatcher(ABC):
    method_name: str
    min_args: int
    max_args: int
    table_arg_index: int
    table_arg_name: str | None = None
    call_context: dict[str, set[str]] | None = None
    session_state: CurrentSessionState | None = None
    is_read: bool | None = None
    is_write: bool | None = None

    def matches(self, node: NodeNG) -> bool:
        return (
            isinstance(node, Call)
            and self._get_table_arg(node) is not None
            and isinstance(node.func, Attribute)
            and Tree(node.func.expr).is_from_module("spark")
        )

    @abstractmethod
    def lint(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterable[Advice]: ...

    @abstractmethod
    def apply(self, from_table: FromTableSqlLinter, index: TableMigrationIndex, node: Call) -> None: ...

    @abstractmethod
    def collect_tables(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterable[UsedTable]: ...

    def _get_table_arg(self, node: Call) -> NodeNG | None:
        node_argc = len(node.args)
        if self.min_args <= node_argc <= self.max_args and self.table_arg_index < node_argc:
            return node.args[self.table_arg_index]
        if not self.table_arg_name:
            return None
        if not node.keywords:
            return None
        for keyword in node.keywords:
            if keyword.arg == self.table_arg_name:
                return keyword.value
        return None

    def _check_call_context(self, node: Call) -> bool:
        assert isinstance(node.func, Attribute)  # Avoid linter warning
        func_name = node.func.attrname
        qualified_name = TreeHelper.get_full_function_name(node)

        # Check if the call_context is None as that means all calls are checked
        if self.call_context is None:
            return True

        # Get the qualified names from the call_context dictionary
        qualified_names = self.call_context.get(func_name)

        # Check if the qualified name is in the set of qualified names that are allowed
        return qualified_name in qualified_names if qualified_names else False


@dataclass
class SparkCallMatcher(_TableNameMatcher):

    def collect_tables(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterable[UsedTable]:
        for used_table in self._collect_tables(from_table, session_state, node):
            if not used_table:
                continue
            yield used_table[1]

    def _collect_tables(
        self,
        from_table: FromTableSqlLinter,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterable[tuple[str, UsedTable] | None]:
        table_arg = self._get_table_arg(node)
        if table_arg is None:
            return
        for inferred in InferredValue.infer_from_node(table_arg, session_state):
            if not inferred.is_inferred():
                yield None
                continue
            table_name = inferred.as_string().strip("'").strip('"')
            info = UsedTable.parse(table_name, from_table.schema, is_read=self.is_read, is_write=self.is_write)
            yield table_name, info

    def lint(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterable[Advice]:
        for used_table in self._collect_tables(from_table, session_state, node):
            if not used_table:
                yield Advisory.from_node(
                    code='cannot-autofix-table-reference',
                    message=f"Can't migrate '{node.as_string()}' because its table name argument cannot be computed",
                    node=node,
                )
                continue
            dst = self._find_dest(index, used_table[1])
            if dst is None:
                continue
            yield Deprecation.from_node(
                code='table-migrated-to-uc',
                message=f"Table {used_table[0]} is migrated to {dst.destination()} in Unity Catalog",
                # SQLGlot does not propagate tokens yet. See https://github.com/tobymao/sqlglot/issues/3159
                node=node,
            )

    def apply(self, from_table: FromTableSqlLinter, index: TableMigrationIndex, node: Call) -> None:
        table_arg = self._get_table_arg(node)
        assert isinstance(table_arg, Const)
        # TODO locate constant when value is inferred
        info = UsedTable.parse(table_arg.value, from_table.schema)
        dst = self._find_dest(index, info)
        if dst is not None:
            table_arg.value = dst.destination()

    @classmethod
    def _find_dest(cls, index: TableMigrationIndex, table: UsedTable) -> TableMigrationStatus | None:
        if table.catalog_name != "hive_metastore":
            return None
        return index.get(table.schema_name, table.table_name)


@dataclass
class ReturnValueMatcher(_TableNameMatcher):

    def matches(self, node: NodeNG) -> bool:
        return (
            isinstance(node, Call) and isinstance(node.func, Attribute) and Tree(node.func.expr).is_from_module("spark")
        )

    def lint(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterator[Advice]:
        assert isinstance(node.func, Attribute)  # always true, avoids a pylint warning
        yield Advisory.from_node(
            code='changed-result-format-in-uc',
            message=f"Call to '{node.func.attrname}' will return a list of <catalog>.<database>.<table> instead of <database>.<table>.",
            node=node,
        )

    def apply(self, from_table: FromTableSqlLinter, index: TableMigrationIndex, node: Call) -> None:
        # No transformations to apply
        return

    def collect_tables(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterable[UsedTable]:
        return []


T = TypeVar("T")


@dataclass
class DirectFilesystemAccessMatcher(_TableNameMatcher):

    def matches(self, node: NodeNG) -> bool:
        return (
            isinstance(node, Call)
            and self._get_table_arg(node) is not None
            and isinstance(node.func, Attribute)
            and (Tree(node.func.expr).is_from_module("spark") or Tree(node.func.expr).is_from_module("dbutils"))
        )

    def lint(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: NodeNG,
    ) -> Iterator[Advice]:
        table_arg = self._get_table_arg(node)
        if table_arg is None:
            return
        for inferred in InferredValue.infer_from_node(table_arg):
            if not inferred.is_inferred():
                logger.debug(f"Could not infer value of {table_arg.as_string()}")
                continue
            value = inferred.as_string()
            if any(pattern.matches(value) for pattern in DIRECT_FS_ACCESS_PATTERNS):
                yield Deprecation.from_node(
                    code='direct-filesystem-access',
                    message=f"The use of direct filesystem references is deprecated: {value}",
                    node=node,
                )

    def apply(self, from_table: FromTableSqlLinter, index: TableMigrationIndex, node: Call) -> None:
        # No transformations to apply
        return

    def collect_tables(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
        node: Call,
    ) -> Iterable[UsedTable]:
        return []  # we don't collect tables through this matcher


class SparkTableNameMatchers:

    def __init__(self, dfsa_matchers_only: bool):

        spark_dfsa_matchers: list[_TableNameMatcher] = [
            DirectFilesystemAccessMatcher(
                "ls", 1, 1, 0, call_context={"ls": {"dbutils.fs.ls"}}, is_read=True, is_write=False
            ),
            DirectFilesystemAccessMatcher(
                "cp", 1, 2, 0, call_context={"cp": {"dbutils.fs.cp"}}, is_read=True, is_write=True
            ),
            DirectFilesystemAccessMatcher("rm", 1, 1, 0, call_context={"rm": {"dbutils.fs.rm"}}, is_write=True),
            DirectFilesystemAccessMatcher(
                "head", 1, 1, 0, call_context={"head": {"dbutils.fs.head"}}, is_read=True, is_write=False
            ),
            DirectFilesystemAccessMatcher(
                "put", 1, 2, 0, call_context={"put": {"dbutils.fs.put"}}, is_read=False, is_write=True
            ),
            DirectFilesystemAccessMatcher(
                "mkdirs", 1, 1, 0, call_context={"mkdirs": {"dbutils.fs.mkdirs"}}, is_read=False, is_write=True
            ),
            DirectFilesystemAccessMatcher(
                "mv", 1, 2, 0, call_context={"mv": {"dbutils.fs.mv"}}, is_read=False, is_write=True
            ),
            DirectFilesystemAccessMatcher("text", 1, 3, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("csv", 1, 1000, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("json", 1, 1000, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("orc", 1, 1000, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("parquet", 1, 1000, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("save", 0, 1000, 0, "path", is_read=False, is_write=True),
            DirectFilesystemAccessMatcher("load", 0, 1000, 0, "path", is_read=True, is_write=False),
            DirectFilesystemAccessMatcher(
                "option", 1, 1000, 1, is_read=True, is_write=False
            ),  # Only .option("path", "xxx://bucket/path") will hit
            DirectFilesystemAccessMatcher("addFile", 1, 3, 0, is_read=False, is_write=True),
            DirectFilesystemAccessMatcher("binaryFiles", 1, 2, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("binaryRecords", 1, 2, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("dump_profiles", 1, 1, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("hadoopFile", 1, 8, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("newAPIHadoopFile", 1, 8, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("pickleFile", 1, 3, 0, is_read=True, is_write=False),
            DirectFilesystemAccessMatcher("saveAsHadoopFile", 1, 8, 0, is_read=False, is_write=True),
            DirectFilesystemAccessMatcher("saveAsNewAPIHadoopFile", 1, 7, 0, is_read=False, is_write=True),
            DirectFilesystemAccessMatcher("saveAsPickleFile", 1, 2, 0, is_read=False, is_write=True),
            DirectFilesystemAccessMatcher("saveAsSequenceFile", 1, 2, 0, is_read=False, is_write=True),
            DirectFilesystemAccessMatcher("saveAsTextFile", 1, 2, 0, is_read=False, is_write=True),
            DirectFilesystemAccessMatcher("load_from_path", 1, 1, 0, is_read=True, is_write=False),
        ]
        if dfsa_matchers_only:
            self._make_matchers(spark_dfsa_matchers)
            return

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.SparkSession.html
        # spark.sql is handled by a dedicated linter
        spark_session_matchers: list[_TableNameMatcher] = [SparkCallMatcher("table", 1, 1, 0)]

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Catalog.html
        spark_catalog_matchers: list[_TableNameMatcher] = [
            SparkCallMatcher("cacheTable", 1, 2, 0, "tableName", is_read=True, is_write=False),
            SparkCallMatcher("createTable", 1, 1000, 0, "tableName", is_read=False, is_write=True),
            SparkCallMatcher("createExternalTable", 1, 1000, 0, "tableName", is_read=False, is_write=True),
            SparkCallMatcher("getTable", 1, 1, 0, is_read=True, is_write=False),
            SparkCallMatcher("isCached", 1, 1, 0, is_read=True, is_write=False),
            SparkCallMatcher("listColumns", 1, 2, 0, "tableName", is_read=True, is_write=False),
            SparkCallMatcher("tableExists", 1, 2, 0, "tableName", is_read=True, is_write=False),
            SparkCallMatcher("recoverPartitions", 1, 1, 0, is_read=True, is_write=False),
            SparkCallMatcher("refreshTable", 1, 1, 0, is_read=True, is_write=False),
            SparkCallMatcher("uncacheTable", 1, 1, 0, is_read=True, is_write=False),
            ReturnValueMatcher("listTables", 0, 2, 0, is_read=True, is_write=False),
        ]

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.html
        spark_dataframe_matchers: list[_TableNameMatcher] = [
            SparkCallMatcher("writeTo", 1, 1, 0, is_read=False, is_write=True),
        ]

        # nothing to migrate in Column, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Column.html
        # nothing to migrate in Observation, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Observation.html
        # nothing to migrate in Row, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Row.html
        # nothing to migrate in GroupedData, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.GroupedData.html
        # nothing to migrate in PandasCogroupedOps, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.PandasCogroupedOps.html
        # nothing to migrate in DataFrameNaFunctions, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameNaFunctions.html
        # nothing to migrate in DataFrameStatFunctions, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameStatFunctions.html
        # nothing to migrate in Window, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Window.html

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.html
        spark_dataframereader_matchers: list[_TableNameMatcher] = [
            SparkCallMatcher(
                "table", 1, 1, 0, is_read=True, is_write=False
            ),  # TODO good example of collision, see spark_session_calls
        ]

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameWriter.html
        spark_dataframewriter_matchers: list[_TableNameMatcher] = [
            SparkCallMatcher("insertInto", 1, 2, 0, "tableName", is_read=False, is_write=True),
            # TODO jdbc: could the url be a databricks url, raise warning ?
            SparkCallMatcher("saveAsTable", 1, 4, 0, "name", is_read=False, is_write=True),
        ]

        # nothing to migrate in DataFrameWriterV2, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameWriterV2.html
        # nothing to migrate in UDFRegistration, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.UDFRegistration.html

        # nothing to migrate in UserDefinedFunction, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.UserDefinedFunction.html
        # nothing to migrate in UserDefinedTableFunction, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.UserDefinedTableFunction.html
        self._make_matchers(
            spark_dfsa_matchers
            + spark_session_matchers
            + spark_catalog_matchers
            + spark_dataframe_matchers
            + spark_dataframereader_matchers
            + spark_dataframewriter_matchers
        )

    def _make_matchers(self, matchers: list[_TableNameMatcher]) -> None:
        self._matchers = {}
        for matcher in matchers:
            self._matchers[matcher.method_name] = matcher

    @property
    def matchers(self) -> dict[str, _TableNameMatcher]:
        return self._matchers


class SparkTableNamePyLinter(PythonLinter, Fixer, TablePyCollector):

    def __init__(
        self,
        from_table: FromTableSqlLinter,
        index: TableMigrationIndex,
        session_state: CurrentSessionState,
    ):
        self._from_table = from_table
        self._index = index
        self._session_state = session_state
        self._spark_matchers = SparkTableNameMatchers(False).matchers

    @property
    def name(self) -> str:
        # this is the same fixer, just in a different language context
        return self._from_table.name

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        for node in tree.walk():
            matcher = self._find_matcher(node)
            if matcher is None:
                continue
            assert isinstance(node, Call)
            yield from matcher.lint(self._from_table, self._index, self._session_state, node)

    def apply(self, code: str) -> str:
        maybe_tree = Tree.maybe_parse(code)
        if not maybe_tree.tree:
            assert maybe_tree.failure is not None
            logger.warning(maybe_tree.failure.message)
            return code
        tree = maybe_tree.tree
        # we won't be doing it like this in production, but for the sake of the example
        for node in tree.walk():
            matcher = self._find_matcher(node)
            if matcher is None:
                continue
            assert isinstance(node, Call)
            matcher.apply(self._from_table, self._index, node)
        return tree.node.as_string()

    def _find_matcher(self, node: NodeNG) -> _TableNameMatcher | None:
        if not isinstance(node, Call):
            return None
        if not isinstance(node.func, Attribute):
            return None
        matcher = self._spark_matchers.get(node.func.attrname, None)
        if matcher is None:
            return None
        return matcher if matcher.matches(node) else None

    def collect_tables_from_tree(self, tree: Tree) -> Iterable[UsedTableNode]:
        for node in tree.walk():
            matcher = self._find_matcher(node)
            if matcher is None:
                continue
            assert isinstance(node, Call)
            for used_table in matcher.collect_tables(self._from_table, self._index, self._session_state, node):
                yield UsedTableNode(used_table, node)


class _SparkSqlAnalyzer:

    @classmethod
    def _visit_call_nodes(cls, tree: Tree) -> Iterable[tuple[Call, NodeNG]]:
        visitor = MatchingVisitor(Call, [("sql", Attribute), ("spark", Name)])
        visitor.visit(tree.node)
        for call_node in visitor.matched_nodes:
            query = TreeHelper.get_arg(call_node, arg_index=0, arg_name="sqlQuery")
            if query is None:
                continue
            yield call_node, query


class SparkSqlPyLinter(_SparkSqlAnalyzer, PythonLinter, Fixer):

    def __init__(self, sql_linter: SqlLinter, sql_fixer: Fixer | None):
        self._sql_linter = sql_linter
        self._sql_fixer = sql_fixer

    @property
    def name(self) -> str:
        return "<none>" if self._sql_fixer is None else self._sql_fixer.name

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        for call_node, query in self._visit_call_nodes(tree):
            for value in InferredValue.infer_from_node(query):
                if not value.is_inferred():
                    yield Advisory.from_node(
                        code="cannot-autofix-table-reference",
                        message=f"Can't migrate table_name argument in '{query.as_string()}' because its value cannot be computed",
                        node=call_node,
                    )
                    continue
                for advice in self._sql_linter.lint(value.as_string()):
                    yield advice.replace_from_node(call_node)

    def apply(self, code: str) -> str:
        if not self._sql_fixer:
            return code
        maybe_tree = Tree.maybe_normalized_parse(code)
        if maybe_tree.failure:
            logger.warning(maybe_tree.failure.message)
            return code
        assert maybe_tree.tree is not None
        tree = maybe_tree.tree
        for _call_node, query in self._visit_call_nodes(tree):
            if not isinstance(query, Const) or not isinstance(query.value, str):
                continue
            # TODO avoid applying same fix multiple times
            # this requires changing 'apply' API in order to check advice fragment location
            new_query = self._sql_fixer.apply(query.value)
            query.value = new_query
        return tree.node.as_string()


class SparkSqlDfsaPyCollector(_SparkSqlAnalyzer, DfsaPyCollector):

    def __init__(self, sql_collector: DfsaSqlCollector):
        self._sql_collector = sql_collector

    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]:
        assert self._sql_collector
        for call_node, query in self._visit_call_nodes(tree):
            for value in InferredValue.infer_from_node(query):
                if not value.is_inferred():
                    continue  # TODO error handling strategy
                for dfsa in self._sql_collector.collect_dfsas(value.as_string()):
                    yield DirectFsAccessNode(dfsa, call_node)


class SparkSqlTablePyCollector(_SparkSqlAnalyzer, TablePyCollector):

    def __init__(self, sql_collector: TableSqlCollector):
        self._sql_collector = sql_collector

    def collect_tables_from_tree(self, tree: Tree) -> Iterable[UsedTableNode]:
        assert self._sql_collector
        for call_node, query in self._visit_call_nodes(tree):
            for value in InferredValue.infer_from_node(query):
                if not value.is_inferred():
                    continue  # TODO error handling strategy
                for table in self._sql_collector.collect_tables(value.as_string()):
                    yield UsedTableNode(table, call_node)



import dataclasses
import logging
from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field
from datetime import datetime, timezone

from databricks.sdk.service.workspace import Language

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.assessment.dashboards import Dashboard, LakeviewDashboardCrawler, RedashDashboardCrawler, Query
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
from databricks.labs.ucx.source_code.base import CurrentSessionState, LineageAtom, UsedTable
from databricks.labs.ucx.source_code.directfs_access import DirectFsAccessCrawler, DirectFsAccess
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.used_table import UsedTablesCrawler

logger = logging.getLogger(__name__)


@dataclass
class QueryProblem:
    dashboard_id: str
    dashboard_parent: str
    dashboard_name: str
    query_id: str
    query_parent: str
    query_name: str
    code: str
    message: str


@dataclass
class _ReportingContext:
    linted_queries: set[str] = field(default_factory=set)
    all_problems: list[QueryProblem] = field(default_factory=list)
    all_dfsas: list[DirectFsAccess] = field(default_factory=list)
    all_tables: list[UsedTable] = field(default_factory=list)


class QueryLinter:

    def __init__(
        self,
        sql_backend: SqlBackend,
        inventory_database: str,
        migration_index: TableMigrationIndex,
        directfs_crawler: DirectFsAccessCrawler,
        used_tables_crawler: UsedTablesCrawler,
        dashboard_crawlers: list[LakeviewDashboardCrawler | RedashDashboardCrawler],
        debug_listing_upper_limit: int | None = None,
    ):
        self._sql_backend = sql_backend
        self._migration_index = migration_index
        self._directfs_crawler = directfs_crawler
        self._used_tables_crawler = used_tables_crawler
        self._dashboard_crawlers = dashboard_crawlers
        self._debug_listing_upper_limit = debug_listing_upper_limit

        self._catalog = "hive_metastore"
        self._schema = inventory_database
        self._table = "query_problems"

    @property
    def _full_name(self) -> str:
        """Generates the full name of the table.

        Returns:
            str: The full table name.
        """
        return f"{self._catalog}.{self._schema}.{self._table}"

    def refresh_report(self) -> None:
        assessment_start = datetime.now(timezone.utc)
        context = _ReportingContext()
        self._lint_dashboards(context)
        self._lint_queries(context)
        assessment_end = datetime.now(timezone.utc)
        self._dump_problems(context.all_problems)
        self._dump_dfsas(context.all_dfsas, assessment_start, assessment_end)
        self._dump_used_tables(context.all_tables, assessment_start, assessment_end)

    def _dump_problems(self, problems: Sequence[QueryProblem]) -> None:
        logger.info(f"Saving {len(problems)} linting problems...")
        self._sql_backend.save_table(
            escape_sql_identifier(self._full_name),
            problems,
            QueryProblem,
            mode='overwrite',
        )

    def _dump_dfsas(
        self,
        dfsas: Sequence[DirectFsAccess],
        assessment_start: datetime,
        assessment_end: datetime,
    ) -> None:
        processed_dfsas = []
        for dfsa in dfsas:
            dfsa = dataclasses.replace(
                dfsa,
                assessment_start_timestamp=assessment_start,
                assessment_end_timestamp=assessment_end,
            )
            processed_dfsas.append(dfsa)
        self._directfs_crawler.dump_all(processed_dfsas)

    def _dump_used_tables(
        self,
        used_tables: Sequence[UsedTable],
        assessment_start: datetime,
        assessment_end: datetime,
    ) -> None:
        processed_tables = []
        for used_table in used_tables:
            used_table = dataclasses.replace(
                used_table,
                assessment_start_timestamp=assessment_start,
                assessment_end_timestamp=assessment_end,
            )
            processed_tables.append(used_table)
        self._used_tables_crawler.dump_all(processed_tables)

    def _lint_dashboards(self, context: _ReportingContext) -> None:
        for dashboard, queries in self._list_dashboards_with_queries():
            logger.info(f"Linting dashboard: {dashboard.name} ({dashboard.id})")
            queries_to_lint = []
            for query in queries:
                if query.id in context.linted_queries:
                    continue
                queries_to_lint.append(query)
                context.linted_queries.add(query.id)
            problems, dfsas, tables = self._lint_dashboard_with_queries(dashboard, queries_to_lint)
            context.all_problems.extend(problems)
            context.all_dfsas.extend(dfsas)
            context.all_tables.extend(tables)

    def _list_dashboards_with_queries(self) -> Iterable[tuple[Dashboard, list[Query]]]:
        for crawler in self._dashboard_crawlers:
            for dashboard in crawler.snapshot():
                yield dashboard, list(crawler.list_queries(dashboard))

    def _lint_queries(self, context: _ReportingContext) -> None:
        for query in self._list_queries():
            if query.id in context.linted_queries:
                continue
            logger.info(f"Linting query: {query.name} ({query.id})")
            context.linted_queries.add(query.id)
            problems = self.lint_query(query)
            context.all_problems.extend(problems)
            dfsas = self.collect_dfsas_from_query("no-dashboard-id", query)
            context.all_dfsas.extend(dfsas)
            tables = self.collect_used_tables_from_query("no-dashboard-id", query)
            context.all_tables.extend(tables)

    def _list_queries(self) -> Iterable[Query]:
        for crawler in self._dashboard_crawlers:
            yield from crawler.list_queries()

    def _lint_dashboard_with_queries(
        self, dashboard: Dashboard, queries: list[Query]
    ) -> tuple[Iterable[QueryProblem], Iterable[DirectFsAccess], Iterable[UsedTable]]:
        query_problems: list[QueryProblem] = []
        query_dfsas: list[DirectFsAccess] = []
        query_tables: list[UsedTable] = []
        for query in queries:
            problems = self.lint_query(query)
            for problem in problems:
                query_problems.append(
                    dataclasses.replace(
                        problem,
                        dashboard_id=dashboard.id,
                        dashboard_parent=dashboard.parent or "PARENT",
                        dashboard_name=dashboard.name or "UNKNOWN",
                    )
                )
            dfsas = self.collect_dfsas_from_query(dashboard.id, query)
            for dfsa in dfsas:
                atom = LineageAtom(
                    object_type="DASHBOARD",
                    object_id=dashboard.id,
                    other={"parent": dashboard.parent or "PARENT", "name": dashboard.name or "UNKNOWN"},
                )
                source_lineage = [atom] + dfsa.source_lineage
                query_dfsas.append(dataclasses.replace(dfsa, source_lineage=source_lineage))
            tables = self.collect_used_tables_from_query(dashboard.id, query)
            for table in tables:
                atom = LineageAtom(
                    object_type="DASHBOARD",
                    object_id=dashboard.id,
                    other={"parent": dashboard.parent or "PARENT", "name": dashboard.name or "UNKNOWN"},
                )
                source_lineage = [atom] + table.source_lineage
                query_tables.append(dataclasses.replace(table, source_lineage=source_lineage))
        return query_problems, query_dfsas, query_tables

    def lint_query(self, query: Query) -> Iterable[QueryProblem]:
        if not query.query:
            return
        ctx = LinterContext(self._migration_index, CurrentSessionState())
        linter = ctx.linter(Language.SQL)
        for advice in linter.lint(query.query):
            yield QueryProblem(
                dashboard_id="",
                dashboard_parent="",
                dashboard_name="",
                query_id=query.id,
                query_parent=query.parent or "PARENT",
                query_name=query.name or "UNKNOWN",
                code=advice.code,
                message=advice.message,
            )

    def collect_dfsas_from_query(self, dashboard_id: str, query: Query) -> Iterable[DirectFsAccess]:
        if not query.query:
            return
        ctx = LinterContext(self._migration_index, CurrentSessionState())
        collector = ctx.dfsa_collector(Language.SQL)
        source_id = f"{dashboard_id}/{query.id}"
        source_lineage = [
            LineageAtom(object_type="QUERY", object_id=source_id, other={"name": query.name or "UNKNOWN"})
        ]
        for dfsa in collector.collect_dfsas(query.query):
            yield dfsa.replace_source(source_id=source_id, source_lineage=source_lineage)

    def collect_used_tables_from_query(self, dashboard_id: str, query: Query) -> Iterable[UsedTable]:
        if not query.query:
            return
        ctx = LinterContext(self._migration_index, CurrentSessionState())
        collector = ctx.tables_collector(Language.SQL)
        source_id = f"{dashboard_id}/{query.id}"
        source_lineage = [
            LineageAtom(object_type="QUERY", object_id=source_id, other={"name": query.name or "UNKNOWN"})
        ]
        for table in collector.collect_tables(query.query):
            yield table.replace_source(source_id=source_id, source_lineage=source_lineage)



from __future__ import annotations

import itertools
import json
import logging
from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field

from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.lsql.lakeview import Dashboard as LsqlLakeviewDashboard, Dataset
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError
from databricks.sdk.service.dashboards import Dashboard as SdkLakeviewDashboard
from databricks.sdk.service.sql import Dashboard as SdkRedashDashboard, LegacyQuery

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.owners import AdministratorLocator, Ownership, WorkspacePathOwnership
from databricks.labs.ucx.framework.utils import escape_sql_identifier


logger = logging.getLogger(__name__)


@dataclass
class Query:
    """UCX representation of a Query.

    Note:
        This class is not persisted into an inventory table. If you decide to persist this class, consider (future)
        differences between Redash and Lakeview queries
    """

    id: str
    """The ID for this query."""

    name: str | None = None
    """The title of this query that appears in list views, widget headings, and on the query page."""

    parent: str | None = None
    """The identifier of the workspace folder containing the object."""

    query: str | None = None
    """The text of the query to be run."""

    catalog: str | None = None
    """The name of the catalog to execute this query in."""

    schema: str | None = None
    """The name of the schema to execute this query in."""

    tags: list[str] = field(default_factory=list)
    """The tags set on this dashboard."""

    @classmethod
    def from_legacy_query(cls, query: LegacyQuery) -> Query:
        """Create query from a :class:LegacyQuery"""
        if not query.id:
            raise ValueError(f"Query id is required: {query}")
        kwargs: dict[str, str | list[str]] = {"id": query.id}
        if query.name:
            kwargs["name"] = query.name
        if query.parent:
            kwargs["parent"] = query.parent
        if query.query:
            kwargs["query"] = query.query
        if query.options and query.options.catalog:
            kwargs["catalog"] = query.options.catalog
        if query.options and query.options.schema:
            kwargs["schema"] = query.options.schema
        if query.tags:
            kwargs["tags"] = query.tags
        return cls(**kwargs)  # type: ignore

    @classmethod
    def from_lakeview_dataset(cls, dataset: Dataset, *, parent: str | None = None) -> Query:
        """Create query from a :class:Dataset"""
        if not dataset.name:
            raise ValueError(f"Dataset name is required: {dataset}")
        kwargs = {"id": dataset.name}
        if dataset.display_name:
            kwargs["name"] = dataset.display_name
        if parent:
            kwargs["parent"] = parent
        if dataset.query:
            kwargs["query"] = dataset.query
        return cls(**kwargs)  # type: ignore


@dataclass
class Dashboard:
    """UCX representation of a dashboard."""

    id: str
    """The ID for this dashboard."""

    name: str | None = None
    """The title of the dashboard that appears in list views and at the top of the dashboard page."""

    parent: str | None = None
    """The identifier of the workspace folder containing the object."""

    query_ids: list[str] = field(default_factory=list)
    """The IDs of the queries referenced by this dashboard."""

    tags: list[str] = field(default_factory=list)
    """The tags set on this dashboard."""

    creator_id: str | None = None
    """The ID of the user who owns the dashboard."""

    @classmethod
    def from_sdk_redash_dashboard(cls, dashboard: SdkRedashDashboard) -> Dashboard:
        assert dashboard.id
        kwargs: dict[str, str | list[str] | None] = {"id": dashboard.id}
        if dashboard.name:
            kwargs["name"] = dashboard.name
        if dashboard.parent:
            kwargs["parent"] = dashboard.parent
        if dashboard.tags:
            kwargs["tags"] = dashboard.tags
        if dashboard.user_id:
            kwargs["creator_id"] = str(dashboard.user_id)
        query_ids = []
        for widget in dashboard.widgets or []:
            if widget.visualization is None:
                continue
            if widget.visualization.query is None:
                continue
            if widget.visualization.query.id is None:
                continue
            query_ids.append(widget.visualization.query.id)
        if query_ids:
            kwargs["query_ids"] = query_ids
        return cls(**kwargs)  # type: ignore

    @classmethod
    def from_sdk_lakeview_dashboard(cls, dashboard: SdkLakeviewDashboard) -> Dashboard:
        assert dashboard.dashboard_id
        kwargs: dict[str, str | list[str] | None] = {"id": dashboard.dashboard_id}
        if dashboard.display_name:
            kwargs["name"] = dashboard.display_name
        if dashboard.parent_path:
            kwargs["parent"] = dashboard.parent_path
        lsql_dashboard = _convert_sdk_to_lsql_lakeview_dashboard(dashboard)
        query_ids = [dataset.name for dataset in lsql_dashboard.datasets]
        if query_ids:
            kwargs["query_ids"] = query_ids
        return cls(**kwargs)  # type: ignore


class RedashDashboardCrawler(CrawlerBase[Dashboard]):
    """Crawler for Redash dashboards."""

    def __init__(
        self,
        ws: WorkspaceClient,
        sql_backend: SqlBackend,
        schema: str,
        *,
        include_dashboard_ids: list[str] | None = None,
        include_query_ids: list[str] | None = None,
        debug_listing_upper_limit: int | None = None,
    ):
        super().__init__(sql_backend, "hive_metastore", schema, "redash_dashboards", Dashboard)
        self._ws = ws
        self._include_dashboard_ids = include_dashboard_ids
        self._include_query_ids = include_query_ids
        self._debug_listing_upper_limit = debug_listing_upper_limit

    def _crawl(self) -> Iterable[Dashboard]:
        dashboards = []
        for sdk_dashboard in self._list_dashboards():
            if sdk_dashboard.id is None:
                continue
            dashboard = Dashboard.from_sdk_redash_dashboard(sdk_dashboard)
            dashboards.append(dashboard)
        return dashboards

    def _list_dashboards(self) -> list[SdkRedashDashboard]:
        if self._include_dashboard_ids is not None:
            return self._get_dashboards(*self._include_dashboard_ids)
        try:
            dashboards_iterator = self._ws.dashboards.list()
        except DatabricksError as e:
            logger.error("Cannot list Redash dashboards", exc_info=e)
            return []
        dashboards: list[SdkRedashDashboard] = []
        # Redash APIs are very slow to paginate, especially for large number of dashboards, so we limit the listing
        # to a small number of items in debug mode for the assessment workflow just to complete.
        while self._debug_listing_upper_limit is None or self._debug_listing_upper_limit > len(dashboards):
            try:
                dashboards.append(next(dashboards_iterator))
            except StopIteration:
                break
            except DatabricksError as e:
                logger.error("Cannot list next Redash dashboards page", exc_info=e)
                break
        return dashboards

    def _get_dashboards(self, *dashboard_ids: str) -> list[SdkRedashDashboard]:
        dashboards = []
        for dashboard_id in dashboard_ids:
            dashboard = self._get_dashboard(dashboard_id)
            if dashboard:
                dashboards.append(dashboard)
        return dashboards

    def _get_dashboard(self, dashboard_id: str) -> SdkRedashDashboard | None:
        try:
            return self._ws.dashboards.get(dashboard_id)
        except DatabricksError as e:
            logger.warning(f"Cannot get Redash dashboard: {dashboard_id}", exc_info=e)
            return None

    def _try_fetch(self) -> Iterable[Dashboard]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Dashboard(*row)

    def list_queries(self, dashboard: Dashboard | None = None) -> Iterator[Query]:
        """List queries.

        Args:
            dashboard (DashboardType | None) : List queries for dashboard. If None, list all queries.
                Defaults to None.

        Note:
            This public method does not adhere to the common crawler layout, still, it is implemented to avoid/postpone
            another crawler for the queries by retrieving the queries every time they are requested.
        """
        for query in self._list_legacy_queries(dashboard):
            yield Query.from_legacy_query(query)

    def _list_legacy_queries(self, dashboard: Dashboard | None = None) -> Iterator[LegacyQuery]:
        """List legacy queries.

        Args:
            dashboard (DashboardType | None) : List queries for dashboard. If None, list all queries.
                Defaults to None.
        """
        if dashboard:
            queries_iterator = self._list_legacy_queries_from_dashboard(dashboard)
        else:
            queries_iterator = self._list_all_legacy_queries()
        # Redash APIs are very slow to paginate, especially for large number of dashboards, so we limit the listing
        # to a small number of items in debug mode for the assessment workflow just to complete.
        counter = itertools.count()
        while self._debug_listing_upper_limit is None or self._debug_listing_upper_limit > next(counter):
            try:
                yield next(queries_iterator)
            except StopIteration:
                break

    def _list_all_legacy_queries(self) -> Iterator[LegacyQuery]:
        """List all queries."""
        if self._include_query_ids is not None:
            yield from self._get_legacy_queries(*self._include_query_ids)
        else:
            try:
                yield from self._ws.queries_legacy.list()
            except DatabricksError as e:
                logger.error("Cannot list Redash queries", exc_info=e)

    def _list_legacy_queries_from_dashboard(self, dashboard: Dashboard) -> Iterator[LegacyQuery]:
        """List queries from dashboard."""
        if self._include_query_ids is not None:
            query_ids = set(dashboard.query_ids) & set(self._include_query_ids)
        else:
            query_ids = set(dashboard.query_ids)
        yield from self._get_legacy_queries(*query_ids)

    def _get_legacy_queries(self, *query_ids: str) -> Iterator[LegacyQuery]:
        """Get a legacy queries."""
        for query_id in query_ids:
            query = self._get_legacy_query(query_id)
            if query:
                yield query

    def _get_legacy_query(self, query_id: str) -> LegacyQuery | None:
        """Get a legacy query."""
        try:
            return self._ws.queries_legacy.get(query_id)
        except DatabricksError as e:
            logger.warning(f"Cannot get Redash query: {query_id}", exc_info=e)
            return None


def _convert_sdk_to_lsql_lakeview_dashboard(dashboard: SdkLakeviewDashboard) -> LsqlLakeviewDashboard:
    """Parse a lsql Lakeview dashboard from an SDK Lakeview dashboard.

    Returns :
        LsqlLakeviewDashboard : The parsed dashboard. If the parsing fails, it is an empty dashboard, i.e. a
            dashboard without datasets and pages.
    """
    lsql_dashboard = LsqlLakeviewDashboard([], [])
    if dashboard.serialized_dashboard is not None:
        try:
            lsql_dashboard = LsqlLakeviewDashboard.from_dict(json.loads(dashboard.serialized_dashboard))
        except (KeyError, ValueError) as e:
            logger.warning(f"Error when parsing Lakeview dashboard: {dashboard.dashboard_id}", exc_info=e)
    return lsql_dashboard


class LakeviewDashboardCrawler(CrawlerBase[Dashboard]):
    """Crawler for Lakeview dashboards."""

    def __init__(
        self,
        ws: WorkspaceClient,
        sql_backend: SqlBackend,
        schema: str,
        *,
        include_dashboard_ids: list[str] | None = None,
        include_query_ids: list[str] | None = None,
    ):
        super().__init__(sql_backend, "hive_metastore", schema, "lakeview_dashboards", Dashboard)
        self._ws = ws
        self._include_dashboard_ids = include_dashboard_ids
        self._include_query_ids = include_query_ids

    def _crawl(self) -> Iterable[Dashboard]:
        dashboards = []
        for sdk_dashboard in self._list_dashboards():
            if sdk_dashboard.dashboard_id is None:
                continue
            dashboard = Dashboard.from_sdk_lakeview_dashboard(sdk_dashboard)
            dashboards.append(dashboard)
        return dashboards

    def _list_dashboards(self) -> list[SdkLakeviewDashboard]:
        if self._include_dashboard_ids is not None:
            return self._get_dashboards(*self._include_dashboard_ids)
        try:
            # If the API listing limit becomes an issue in testing, please see the `:class:RedashDashboardCrawler`
            # for an example on how to implement a (debug) rate limit
            return list(self._ws.lakeview.list())  # TODO: Add dashboard summary view?
        except DatabricksError as e:
            logger.error("Cannot list Lakeview dashboards", exc_info=e)
            return []

    def _get_dashboards(self, *dashboard_ids: str) -> list[SdkLakeviewDashboard]:
        dashboards = []
        for dashboard_id in dashboard_ids:
            dashboard = self._get_dashboard(dashboard_id)
            if dashboard:
                dashboards.append(dashboard)
        return dashboards

    def _get_dashboard(self, dashboard_id: str) -> SdkLakeviewDashboard | None:
        try:
            return self._ws.lakeview.get(dashboard_id)
        except DatabricksError as e:
            logger.warning(f"Cannot get Lakeview dashboard: {dashboard_id}", exc_info=e)
            return None

    def _try_fetch(self) -> Iterable[Dashboard]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Dashboard(*row)

    def list_queries(self, dashboard: Dashboard | None = None) -> Iterator[Query]:
        """List queries.

        Args:
            dashboard (DashboardType | None) : List queries for dashboard. If None, list all queries.
                Defaults to None.

        Note:
            This public method does not adhere to the common crawler layout, still, it is implemented to avoid/postpone
            another crawler for the queries by retrieving the queries every time they are requested.

            Different to the Redash crawler, Lakeview queries are part of the (serialized) dashboard definition.
        """
        if dashboard:
            sdk_dashboard = self._get_dashboard(dashboard_id=dashboard.id)
            sdk_dashboards = [sdk_dashboard] if sdk_dashboard else []
        else:
            sdk_dashboards = self._list_dashboards()
        for sdk_dashboard in sdk_dashboards:
            lsql_dashboard = _convert_sdk_to_lsql_lakeview_dashboard(sdk_dashboard)
            for dataset in lsql_dashboard.datasets:
                if self._include_query_ids is not None and dataset.name not in self._include_query_ids:
                    continue
                yield Query.from_lakeview_dataset(dataset, parent=sdk_dashboard.dashboard_id)


class DashboardOwnership(Ownership[Dashboard]):
    """Determine ownership of dashboard in the inventory.

    This is the dashboard creator (if known) otherwise the parent (path) owner (if known).
    """

    def __init__(
        self,
        administrator_locator: AdministratorLocator,
        ws: WorkspaceClient,
        workspace_path_ownership: WorkspacePathOwnership,
    ) -> None:
        super().__init__(administrator_locator)
        self._ws = ws
        self._workspace_path_ownership = workspace_path_ownership

    def _maybe_direct_owner(self, record: Dashboard) -> str | None:
        if record.creator_id:
            creator_name = self._get_user_name(record.creator_id)
            if creator_name:
                return creator_name
        if record.parent:
            return self._workspace_path_ownership.owner_of_path(record.parent)
        return None

    def _get_user_name(self, user_id: str) -> str | None:
        try:
            user = self._ws.users.get(user_id)
            return user.display_name or user.user_name
        except DatabricksError as e:
            logger.warning(f"Could not retrieve user: {user_id}", exc_info=e)
            return None
